import inspect
import warnings
from typing import List, Mapping, Optional, Tuple, Union

import numpy as np
import scipy.stats as st
import xarray as xr

from . import deterministic as dm

XArray = Union[xr.Dataset, xr.DataArray]


def sign_test(
    forecasts1,
    forecasts2,
    observations=None,
    time_dim="time",
    dim=[],
    alpha=0.05,
    metric=None,
    orientation="negative",
):
    """
    Returns the Delsole and Tippett sign test over the given time dimension.

    The sign test can be applied to a wide class of measures of forecast quality,
    including ordered (ranked) categorical data. It is independent of
    distributional assumptions about the forecast errors. This is different than
    alternative measures like correlation and mean square error, which assume that
    the metrics were computed from independent samples. However, skill metrics
    computed over a common period with a common set of observations are not
    independent. For example, different forecasts tend to bust for the same event.
    This procedure is equivalent to testing whether a coin is fair based on the
    frequency of heads. The null hypothesis is that the difference between the
    median scores is zero.

    Parameters
    ----------
    forecasts1 : xarray.Dataset or xarray.DataArray
        forecasts1 to be compared to observations
    forecasts2 : xarray.Dataset or xarray.DataArray
        forecasts2 to be compared to observations
    observations : xarray.Dataset or xarray.DataArray or None
        observation to be compared to both forecasts. Only used if ``metric`` is
        provided, otherwise it is assumed that both forecasts have already been
        compared to observations and this input is ignored. Please
        adjust ``orientation`` accordingly. Defaults to None.
    time_dim : str
        time dimension of dimension over which to compute the random walk.
        This dimension is not reduced, unlike in other xskillscore functions.
        Defaults to ``'time'``.
    dim : str or list of str
        dimensions to apply metric to if ``metric`` is provided. Cannot contain
        ``time_dim``. Ignored if ``metric`` is None. Defaults to [].
    alpha : float
        significance level for random walk.
    metric : callable, str, optional
        metric to compare forecast# with observations if ``metric`` is not None. If
        ``metric`` is None, assume that forecast# have been compared observations
        before using ``sign_test``. Make sure to adjust ``orientation`` if
        ``metric`` is None. Use ``metric=categorical``, if the winning forecast
        should only be rewarded a point if it exactly equals the observations. Also
        allows strings to be convered to ``xskillscore.{metric}``. Defaults to None.
    orientation : str
        One of [``'positive'``, ``'negative'``]. Which skill values correspond to
        better skill? Smaller values (``'negative'``) or larger values
        (``'positive'``)? Defaults to ``'negative'``.
        Ignored if ``metric== categorical``.

    Returns
    -------
    xarray.DataArray or xarray.Dataset
        boolean whether ``forecast1`` is significantly different to ``forecast2``.
    xarray.DataArray or xarray.Dataset
        walk values shows how often ``forecast1`` is better ``forecast2``.
    xarray.DataArray or xarray.Dataset
        confidence boundary for a random walk at significance level ``alpha``.

    Examples
    --------
    >>> f1 = xr.DataArray(np.random.normal(size=(30)),
    ...                   coords=[('time', np.arange(30))])
    >>> f2 = f1 + 2
    >>> o = xr.DataArray(np.random.normal(size=(30)),
    ...                  coords=[('time', np.arange(30))])
    >>> significantly_different, walk, confidence = xs.sign_test(
    ... f1, f2, o, time_dim='time', metric='mae', orientation='negative'
    ... )
    >>> walk.plot() # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> confidence.plot(color='gray') # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> (-1 * confidence).plot(color='gray') # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> walk
    <xarray.DataArray (time: 30)>
    array([ 1,  0,  1,  2,  1,  2,  3,  4,  5,  6,  5,  6,  7,  6,  7,  8,  9,
           10,  9, 10, 11, 12, 13, 12, 11, 12, 13, 14, 15, 14])
    Coordinates:
      * time     (time) int64 0 1 2 3 4 5 6 7 8 9 ... 20 21 22 23 24 25 26 27 28 29
    >>> significantly_different
    <xarray.DataArray (time: 30)>
    array([False, False, False, False, False, False, False, False, False,
           False, False, False, False, False, False,  True,  True,  True,
            True,  True,  True,  True,  True,  True,  True,  True,  True,
            True,  True,  True])
    Coordinates:
      * time     (time) int64 0 1 2 3 4 5 6 7 8 9 ... 20 21 22 23 24 25 26 27 28 29
        alpha    float64 0.05

    References
    ----------
        * DelSole, T., & Tippett, M. K. (2016). Forecast Comparison Based on Random
          Walks. Monthly Weather Review, 144(2), 615–626. doi: 10/f782pf
    """

    def _categorical_metric(observations, forecasts, dim):
        """Returns True where forecasts exactly equals observations"""
        return observations == forecasts

    if orientation not in ["negative", "positive"]:
        raise ValueError(
            '`orientation` requires to be either "positive" or'
            f'"negative"], found {orientation}.'
        )

    if isinstance(dim, str):
        dim = [dim]
    if time_dim in dim:
        raise ValueError("`dim` cannot contain `time_dim`")

    if metric is not None:
        # make sure metric is a callable
        if isinstance(metric, str):
            if metric == "categorical":
                metric = _categorical_metric
                if orientation != "positive":
                    warnings.warn(
                        'Changing to "orientation=positive" \
                            for consistency with "metric=categorical"',
                        UserWarning,
                    )
                orientation = "positive"
            else:
                import xskillscore as xs

                if hasattr(xs, metric):
                    metric = getattr(xs, metric)
                else:
                    raise ValueError(f"xskillscore.{metric} does not exist.")
        elif not callable(metric):
            raise ValueError(
                f'metric needs to be a function/callable, string ["categorical", '
                f"xskillscore.{{{metric}}}] or None, found {type(metric)}"
            )
        if observations is not None:
            # Compare the forecasts and observations using metric
            metric_f1o = metric(observations, forecasts1, dim=dim)
            metric_f2o = metric(observations, forecasts2, dim=dim)
        else:
            raise ValueError("observations must be provided when metric is provided")

    else:  # if metric=None, already evaluated
        if observations is not None:
            warnings.warn(
                "Ignoring provided observations because no metric was provided",
                UserWarning,
            )
        metric_f1o = forecasts1
        metric_f2o = forecasts2

    # Adjust for orientation of metric
    if orientation == "positive":
        if metric == _categorical_metric:
            metric_f1o = ~metric_f1o
            metric_f2o = ~metric_f2o
        else:
            metric_f1o = -metric_f1o
            metric_f2o = -metric_f2o

    walk = (1 * (metric_f1o < metric_f2o) - 1 * (metric_f2o < metric_f1o)).cumsum(
        time_dim
    )

    # Estimate 1 - alpha confidence interval -----
    notnan = 1 * (metric_f1o.notnull() & metric_f2o.notnull())
    N = notnan.cumsum(time_dim)
    # z_alpha is the value at which the standardized cumulative Gaussian distributed
    # exceeds alpha
    confidence = st.norm.ppf(1 - alpha / 2) * np.sqrt(N)
    confidence.coords["alpha"] = alpha
    significantly_different = np.abs(walk) > confidence
    return significantly_different, walk, confidence


def halfwidth_ci_test(
    forecasts1: XArray,
    forecasts2: XArray,
    observations: Optional[XArray] = None,
    metric: Optional[str] = None,
    dim: Optional[Union[str, List[str]]] = None,
    time_dim: str = "time",
    alpha: float = 0.05,
    **kwargs: Mapping,
) -> Tuple[XArray, XArray, XArray]:
    """
    Returns the Jolliffe and Ebert significance test.

    Tests whether forecasts1 and forecasts2 have different distance from
    observations at significance level alpha.
    https://www.cawcr.gov.au/projects/verification/CIdiff/FAQ-CIdiff.html


    .. note::
        ``alpha`` is the desired significance level and the maximum acceptable risk of
        falsely rejecting the null-hypothesis. The smaller the value of α the greater
        the strength of the test. The confidence level of the test is defined as
        1 - alpha, and often expressed as a percentage. So for example a significance
        level of 0.05, is equivalent to a 95% confidence level.
        Source: NIST/SEMATECH e-Handbook of Statistical Methods.
        https://www.itl.nist.gov/div898/handbook/prc/section1/prc14.htm

    Parameters
    ----------
    forecasts1 : xarray.Dataset or xarray.DataArray
        first forecast to be compared to the observations.
    forecasts2 : xarray.Dataset or xarray.DataArray
        second forecast to be compared to the observations.
    observations : xarray.Dataset or xarray.DataArray, optional
        observations to be compared to both forecasts. if None, assumes that arguments
        forecasts1 and forecasts2 are already MAEs. Defaults to None.
    metric : str, optional
        Name of distance metric function to be used for computing the error between
        forecasts and observation. It can be any of the xskillscore distance metric
        function except for ``mape``. Valid metrics are ``me``, ``rmse``, ``mse``,
        ``mae``, ``median_absolute_error`` and ``smape``. Note that if metric is None,
        observations must also be None.
        Defaults to None.
    time_dim : str, optional
        time dimension of dimension over which to compute the temporal correlation.
        Defaults to ``'time'``.
    dim : str or list of str, optional
        dimensions to apply metric function to. Cannot contain ``time_dim``.
        Defaults to None which is then converted to ``[]`` since ``dim=None`` must not
        be passed to metric functions.
    alpha : float, optional
        significance level alpha that forecast1 is different than forecast2.
    **kwargs : dict, optional
        Optional keyword arguments passed directly on to call ``metric``,
        excluding ``dim``.

    Returns
    -------
    xarray.DataArray or xarray.Dataset
        boolean whether the difference in scores (score(f2) - score(f1)) are significant.
    xarray.DataArray or xarray.Dataset
        difference in scores (score(f2) - score(f1)) reduced by ``dim`` and ``time_dim``.
    xarray.DataArray or xarray.Dataset
        half-width of the confidence interval at the significance level ``alpha``.

    Examples
    --------
    >>> f1 = xr.DataArray(np.random.normal(size=(30)),
    ...                   coords=[('time', np.arange(30))])
    >>> f2 = xr.DataArray(np.random.normal(size=(30)),
    ...                   coords=[('time', np.arange(30))])
    >>> o = xr.DataArray(np.random.normal(size=(30)),
    ...                  coords=[('time', np.arange(30))])
    >>> significantly_different, diff, hwci = xs.halfwidth_ci_test(
    ...    f1, f2, o, "mae", time_dim='time', dim=[], alpha=0.05
    ... )
    >>> significantly_different
    <xarray.DataArray ()>
    array(False)
    >>> diff
    <xarray.DataArray ()>
    array(-0.01919449)
    >>> hwci
    <xarray.DataArray ()>
    array(0.38729387)
    >>> # absolute magnitude of difference is smaller than half-width of
    >>> # confidence interval, therefore not significant at level alpha=0.05
    >>> # now comparing against an offset f2, the difference in MAE is significant
    >>> significantly_different, diff, hwci = xs.halfwidth_ci_test(
    ... f1, f2 + 2., o, "mae", time_dim='time', dim=[], alpha=0.05
    ... )
    >>> significantly_different
    <xarray.DataArray ()>
    array(True)

    References
    ----------
        * https://www.cawcr.gov.au/projects/verification/CIdiff/FAQ-CIdiff.html
    """
    if isinstance(dim, str):
        dim = [dim]
    elif dim is None:
        dim = []

    if time_dim in dim:
        raise ValueError("`dim` cannot contain `time_dim`")

    msg = f"`alpha` must be between 0 and 1 or `return_p`, found {alpha}."
    if isinstance(alpha, (str, int)) and alpha != "return_p":
        raise ValueError(msg)

    if isinstance(alpha, float) and not (0 < alpha < 1):
        raise ValueError(msg)

    if observations is not None and isinstance(metric, str):
        valid_metrics = [
            "me",
            "rmse",
            "mse",
            "mae",
            "median_absolute_error",
            "smape",
        ]
        if metric not in valid_metrics:
            msg = (
                f"`metric` should be a valid distance metric function, found {metric}."
                " Valid metrics are:\n"
                ", ".join(valid_metrics)
            )
            raise ValueError(msg)

        err_func = getattr(dm, metric)
        if dim is not None and "dim" in kwargs:
            kwargs.pop("dim")

        params = inspect.signature(err_func).parameters
        missing_args = [
            p
            for p, v in params.items()
            if v.default == inspect._empty and p not in kwargs and p not in ["a", "b"]  # type: ignore
        ]
        if len(missing_args) > 0:
            msg = (
                f"The following positional arguments for {metric} are missing:\n"
                ", ".join(missing_args)
            )
            raise ValueError(msg)

        # Compare the forecasts and observations using metric
        score_f1o = err_func(observations, forecasts1, dim=dim, **kwargs)
        score_f2o = err_func(observations, forecasts2, dim=dim, **kwargs)
    elif observations is None and metric is None:
        score_f1o = forecasts1
        score_f2o = forecasts2
    else:
        msg = (
            "Both `metric` and `observations` arguments must be either None or "
            "valid inputs."
        )
        raise ValueError(msg)

    pearson_r_f1f2 = dm.pearson_r(score_f1o, score_f2o, dim=time_dim)

    # diff metric
    diff = score_f2o.mean(time_dim) - score_f1o.mean(time_dim)

    notnan = 1 * (score_f1o.notnull() & score_f2o.notnull())
    N = notnan.sum(time_dim)
    # average variances and take square root instead of averaging standard deviations
    std = ((score_f1o.var(time_dim) + score_f2o.var(time_dim)) * 0.5) ** 0.5

    confidence = st.norm.ppf(1.0 - 0.5 * alpha)
    # half width of the confidence interval
    hwci = (2 * (1 - pearson_r_f1f2)) ** 0.5 * confidence * std / N ** 0.5

    significantly_different = np.abs(diff) > hwci  # metric difference significant?
    return significantly_different, diff, hwci



# flake8: noqa
from pkg_resources import DistributionNotFound, get_distribution

from .core import resampling
from .core.accessor import XSkillScoreAccessor
from .core.comparative import halfwidth_ci_test, sign_test
from .core.contingency import Contingency
from .core.deterministic import (
    effective_sample_size,
    linslope,
    mae,
    mape,
    me,
    median_absolute_error,
    mse,
    pearson_r,
    pearson_r_eff_p_value,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_eff_p_value,
    spearman_r_p_value,
)
from .core.probabilistic import (
    brier_score,
    crps_ensemble,
    crps_gaussian,
    crps_quadrature,
    discrimination,
    rank_histogram,
    reliability,
    roc,
    rps,
    threshold_brier_score,
)
from .core.resampling import resample_iterations, resample_iterations_idx
from .versioning.print_versions import show_versions

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass



import numpy as np
import numpy.testing as npt
import pytest
import xarray as xr
from sklearn.metrics import confusion_matrix, roc_auc_score, roc_curve

from xskillscore import Contingency, roc

DIMS = (["time"], ["lon"], ["lat"], "time", ["lon", "lat", "time"])
CATEGORY_EDGES = [
    np.array([-np.inf, 0.5, np.inf]),
    np.array([-np.inf, 0.5, 1.5, np.inf]),
]


@pytest.mark.parametrize("type", ["da", "ds", "chunked_da", "chunked_ds"])
@pytest.mark.parametrize("dim", DIMS)
@pytest.mark.parametrize("category_edges", CATEGORY_EDGES)
def test_Contingency_table(
    observation_3d_int, forecast_3d_int, category_edges, dim, type
):
    """Test that contingency table builds successfully"""
    if "ds" in type:
        name = "var"
        observation_3d_int = observation_3d_int.to_dataset(name=name)
        forecast_3d_int = forecast_3d_int.to_dataset(name=name)
    if "chunked" in type:
        observation_3d_int = observation_3d_int.chunk()
        forecast_3d_int = forecast_3d_int.chunk()
    cont_table = Contingency(
        observation_3d_int, forecast_3d_int, category_edges, category_edges, dim=dim
    )
    assert cont_table


@pytest.mark.parametrize("category_edges", CATEGORY_EDGES)
def test_Contingency_table_values(observation_3d_int, forecast_3d_int, category_edges):
    """Test contingency table values against sklearn.metrics.confusion_matrix
    for 1D data"""

    def logical(ds, edges):
        """Convert input xarray DataArray or Dataset to integers corresponding
        to which bin the data falls in.
        In first bin -> 0; in second bin -> 1; in third bin -> 2; etc
        """
        ds_out = 0 * ds.copy()
        for i in range(1, len(edges) - 1):
            ds_out += i * ((ds > edges[i]) & (ds <= edges[i + 1]))
        return ds_out

    cont_table = Contingency(
        observation_3d_int, forecast_3d_int, category_edges, category_edges, dim="time"
    )
    for lon in forecast_3d_int.lon:
        for lat in forecast_3d_int.lat:
            observation_1d = logical(
                observation_3d_int.sel(lon=lon, lat=lat), category_edges
            )
            forecast_1d = logical(forecast_3d_int.sel(lon=lon, lat=lat), category_edges)
            sklearn_cont_table_1d = confusion_matrix(
                observation_1d, forecast_1d, labels=range(len(category_edges) - 1)
            )
            xs_cont_table_1d = (
                cont_table.table.sel(lon=lon, lat=lat)
                .transpose("observations_category", "forecasts_category")
                .values
            )
            npt.assert_allclose(sklearn_cont_table_1d, xs_cont_table_1d)


@pytest.mark.parametrize(
    "method, expected",
    [
        ("hits", 2),
        ("misses", 2),
        ("false_alarms", 1),
        ("correct_negatives", 2),
        ("bias_score", 3 / 4),
        ("hit_rate", 1 / 2),
        ("false_alarm_ratio", 1 / 3),
        ("false_alarm_rate", 1 / 3),
        ("success_ratio", 2 / 3),
        ("threat_score", 2 / 5),
        ("equit_threat_score", (2 - 12 / 7) / (5 - 12 / 7)),
        ("odds_ratio", 2),
        ("odds_ratio_skill_score", 1 / 3),
        ("accuracy", 4 / 7),
        ("heidke_score", (4 - 24 / 7) / (7 - 24 / 7)),
        ("peirce_score", 1 / 2 - 1 / 3),
        ("gerrity_score", (2 * (3 / 4) + 1 * -1 + 2 * -1 + 2 * (4 / 3)) / 7),
    ],
)
def test_dichotomous_scores(dichotomous_Contingency_1d, method, expected):
    """Test score for simple 2x2 contingency table against hand-computed values
    Scores are for H/TP: 2, M/FN: 2, FA/FP: 1, CN/TN: 2
    """
    xs_score = getattr(dichotomous_Contingency_1d, method)().item()
    npt.assert_almost_equal(xs_score, expected)



import xarray as xr

from .deterministic import (
    effective_sample_size,
    linslope,
    mae,
    mape,
    me,
    median_absolute_error,
    mse,
    pearson_r,
    pearson_r_eff_p_value,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_eff_p_value,
    spearman_r_p_value,
)
from .probabilistic import (
    brier_score,
    crps_ensemble,
    crps_gaussian,
    crps_quadrature,
    discrimination,
    rank_histogram,
    reliability,
    roc,
    rps,
    threshold_brier_score,
)


@xr.register_dataset_accessor("xs")
class XSkillScoreAccessor(object):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def _in_ds(self, x):
        """
        If x is not a string, presumably an array, return the array.
        Else x is a string, presumably within ds, return the ds variable.
        """
        if not isinstance(x, str):
            return x
        else:
            return self._obj[x]

    def linslope(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return linslope(a, b, *args, **kwargs)

    def pearson_r(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return pearson_r(a, b, *args, **kwargs)

    def r2(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return r2(a, b, *args, **kwargs)

    def pearson_r_p_value(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return pearson_r_p_value(a, b, *args, **kwargs)

    def effective_sample_size(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return effective_sample_size(a, b, *args, **kwargs)

    def pearson_r_eff_p_value(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return pearson_r_eff_p_value(a, b, *args, **kwargs)

    def spearman_r(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return spearman_r(a, b, *args, **kwargs)

    def spearman_r_p_value(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return spearman_r_p_value(a, b, *args, **kwargs)

    def spearman_r_eff_p_value(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return spearman_r_eff_p_value(a, b, *args, **kwargs)

    def me(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return me(a, b, *args, **kwargs)

    def rmse(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return rmse(a, b, *args, **kwargs)

    def mse(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return mse(a, b, *args, **kwargs)

    def mae(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return mae(a, b, *args, **kwargs)

    def median_absolute_error(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return median_absolute_error(a, b, *args, **kwargs)

    def mape(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return mape(a, b, *args, **kwargs)

    def smape(self, a, b, *args, **kwargs):
        a = self._in_ds(a)
        b = self._in_ds(b)
        return smape(a, b, *args, **kwargs)

    def crps_gaussian(self, observations, mu, sig, *args, **kwargs):
        observations = self._in_ds(observations)
        mu = self._in_ds(mu)
        sig = self._in_ds(sig)
        return crps_gaussian(observations, mu, sig, *args, **kwargs)

    def crps_ensemble(self, observations, forecasts, *args, **kwargs):
        observations = self._in_ds(observations)
        forecasts = self._in_ds(forecasts)
        return crps_ensemble(observations, forecasts, *args, **kwargs)

    def crps_quadrature(self, x, cdf_or_dist, *args, **kwargs):
        x = self._in_ds(x)
        return crps_quadrature(x, cdf_or_dist, *args, **kwargs)

    def threshold_brier_score(
        self, observations, forecasts, threshold, *args, **kwargs
    ):
        observations = self._in_ds(observations)
        forecasts = self._in_ds(forecasts)
        threshold = self._in_ds(threshold)
        return threshold_brier_score(
            observations, forecasts, threshold, *args, **kwargs
        )

    def brier_score(self, observations, forecasts, *args, **kwargs):
        observations = self._in_ds(observations)
        forecasts = self._in_ds(forecasts)
        return brier_score(observations, forecasts, *args, **kwargs)

    def rps(self, observations, forecasts, *args, **kwargs):
        observations = self._in_ds(observations)
        forecasts = self._in_ds(forecasts)
        return rps(observations, forecasts, *args, **kwargs)

    def rank_histogram(self, observations, forecasts, *args, **kwargs):
        observations = self._in_ds(observations)
        forecasts = self._in_ds(forecasts)
        return rank_histogram(observations, forecasts, *args, **kwargs)

    def discrimination(self, observations, forecasts, *args, **kwargs):
        observations = self._in_ds(observations)
        forecasts = self._in_ds(forecasts)
        return discrimination(observations, forecasts, *args, **kwargs)

    def reliability(self, observations, forecasts, *args, **kwargs):
        observations = self._in_ds(observations)
        forecasts = self._in_ds(forecasts)
        return reliability(observations, forecasts, *args, **kwargs)

    def roc(self, observations, forecasts, *args, **kwargs):
        observations = self._in_ds(observations)
        forecasts = self._in_ds(forecasts)
        return roc(observations, forecasts, *args, **kwargs)



from typing import Callable, List

import numpy as np
import pytest

from xskillscore.core.deterministic import (
    linslope,
    mae,
    mape,
    median_absolute_error,
    mse,
    pearson_r,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_p_value,
)

# Should only have masking issues when pulling in masked
# grid cells over space.
AXES = ("time", "lat", "lon", ["lat", "lon"], ["time", "lat", "lon"])

distance_metrics: List[Callable] = [mae, mape, median_absolute_error, mse, rmse, smape]
correlation_metrics: List[Callable] = [
    linslope,
    pearson_r,
    pearson_r_p_value,
    r2,
    spearman_r,
    spearman_r_p_value,
]


@pytest.mark.parametrize("metric", correlation_metrics + distance_metrics)
@pytest.mark.parametrize("dim", AXES)
def test_metrics_masked(a_fixed_nan, b_fixed_nan, dim, metric):
    """Test for all distance-based metrics whether result of skipna does not
    contain any nans when applied along dim with nans."""
    a = a_fixed_nan
    b = b_fixed_nan
    res_skipna = metric(a, b, dim, skipna=True)
    res_no_skipna = metric(a, b, dim, skipna=False)

    if "lon" in dim or "lat" in dim:  # metric is applied along axis with nans
        # res_skipna shouldnt have nans
        if metric not in [spearman_r_p_value, pearson_r_p_value]:
            assert not np.isnan(res_skipna).any()
        # res_no_skipna should have different result then skipna
        assert (res_no_skipna != res_skipna).any()
    else:  # metric is applied along axis without nans
        res_skipna_where_masked = res_skipna.isel(lon=[1, 2], lat=[1, 2])
        res_no_skipna_where_masked = res_no_skipna.isel(lon=[1, 2], lat=[1, 2])

        assert np.isnan(res_skipna_where_masked).all()
        assert np.isnan(res_no_skipna_where_masked).all()
        # res_skipna should have a few nans
        assert np.isnan(res_skipna).any()
        # res_no_skipna should have a few nans
        assert np.isnan(res_no_skipna).any()
        # # res_no_skipna should have different result then skipna
        assert (res_no_skipna != res_skipna).any()



from xskillscore.core.deterministic import (
    effective_sample_size,
    pearson_r_eff_p_value,
    pearson_r_p_value,
    spearman_r_eff_p_value,
    spearman_r_p_value,
)

DIM = "time"


def test_eff_sample_size_smaller_than_n(a, b):
    """Tests that the effective sample size is less than or equal to the normal
    sample size."""
    # Can just use `.size` here since we don't
    # have any NaNs in the test.
    N = a[DIM].size
    eff_N = effective_sample_size(a, b, DIM)
    assert (eff_N <= N).all()


def test_eff_pearson_p_greater_or_equal_to_normal_p(a, b):
    """Tests that the effective Pearson p value is greater than or equal to the
    normal Pearson p value."""
    print(a)
    print(b)
    normal_p = pearson_r_p_value(a, b, DIM)
    eff_p = pearson_r_eff_p_value(a, b, DIM)
    print("\n" * 3)
    print(normal_p)
    print(eff_p)
    assert (eff_p >= normal_p).all()


def test_eff_spearman_p_greater_or_equal_to_normal_p(a, b):
    """Tests that the effective Spearman p value is greater than or equal to the
    normal Spearman p value."""
    normal_p = spearman_r_p_value(a, b, DIM)
    eff_p = spearman_r_eff_p_value(a, b, DIM)
    assert (eff_p >= normal_p).all()



from functools import wraps

import numpy as np
import xarray as xr

from .utils import histogram

__all__ = ["Contingency"]

OBSERVATIONS_NAME = "observations"
FORECASTS_NAME = "forecasts"


def _get_category_bounds(category_edges):
    """Return formatted string of category bounds given list of category edges"""
    bounds = [
        f"[{str(category_edges[i])}, {str(category_edges[i + 1])})"
        for i in range(len(category_edges) - 2)
    ]
    # Last category is right edge inclusive
    bounds.append(f"[{str(category_edges[-2])}, {str(category_edges[-1])}]")
    return bounds


def dichotomous_only(method):
    """Decorator for methods that are defined for dichotomous forecasts only"""

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.dichotomous:
            raise AttributeError(
                f"{method.__name__} can only be computed for \
                    dichotomous (2-category) data"
            )
        return method(self, *args, **kwargs)

    return wrapper


def _display_metadata(self):
    """Called when Contingency objects are printed"""
    header = f"<xskillscore.{type(self).__name__}>\n"
    summary = header + "\n".join(str(self.table).split("\n")[1:]) + "\n"
    return summary


class Contingency:
    """Class for contingency based skill scores

    Parameters
    ----------
    observations : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    forecasts : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    observation_category_edges : array_like
        Bin edges for categorising observations. Similar to np.histogram, \
        all but the last (righthand-most) bin include the left edge and \
        exclude the right edge. The last bin includes both edges.
    forecast_category_edges : array_like
        Bin edges for categorising forecasts. Similar to np.histogram, \
        all but the last (righthand-most) bin include the left edge and \
        exclude the right edge. The last bin includes both edges.
    dim : str, list
        The dimension(s) over which to compute the contingency table

    Returns
    -------
    xskillscore.Contingency

    Examples
    --------
    >>> da = xr.DataArray(np.random.normal(size=(3, 3)),
    ...                   coords=[("x", np.arange(3)), ("y", np.arange(3))])
    >>> o = xr.Dataset({"var1": da, "var2": da})
    >>> f = o * 1.1
    >>> o_category_edges = np.linspace(-2, 2, 5)
    >>> f_category_edges = np.linspace(-3, 3, 5)
    >>> xs.Contingency(o, f,
    ...                o_category_edges, f_category_edges,
    ...                dim=['x', 'y']) # doctest: +SKIP
    <xskillscore.Contingency>
    Dimensions:                       (forecasts_category: 4, observations_category: 4)
    Coordinates:
        observations_category_bounds  (observations_category) <U12 '[-2.0, -1.0)'...
        forecasts_category_bounds     (forecasts_category) <U12 '[-3.0, -1.5)' .....
      * observations_category         (observations_category) int64 1 2 3 4
      * forecasts_category            (forecasts_category) int64 1 2 3 4
    Data variables:
        var1                           (observations_category, forecasts_category) int64
        var2                           (observations_category, forecasts_category) int64


    References
    ----------
    http://www.cawcr.gov.au/projects/verification/
    """

    def __init__(
        self,
        observations,
        forecasts,
        observation_category_edges,
        forecast_category_edges,
        dim,
    ):
        self._observations = observations.copy()
        self._forecasts = forecasts.copy()
        self._observation_category_edges = observation_category_edges.copy()
        self._forecast_category_edges = forecast_category_edges.copy()
        self._dichotomous = (
            True
            if (len(observation_category_edges) - 1 == 2)
            & (len(forecast_category_edges) - 1 == 2)
            else False
        )
        self._table = self._get_contingency_table(dim)

    @property
    def observations(self):
        return self._observations

    @property
    def forecasts(self):
        return self._forecasts

    @property
    def observation_category_edges(self):
        return self._observation_category_edges

    @property
    def forecast_category_edges(self):
        return self._forecast_category_edges

    @property
    def dichotomous(self):
        return self._dichotomous

    @property
    def table(self):
        return self._table

    def _get_contingency_table(self, dim):
        """Build the contingency table

        Parameters
        ----------
        dim : str, list
            The dimension(s) over which to compute the contingency table

        Returns
        -------
        xarray.Dataset or xarray.DataArray
        """

        table = histogram(
            self.observations,
            self.forecasts,
            bins=[self.observation_category_edges, self.forecast_category_edges],
            bin_names=[OBSERVATIONS_NAME, FORECASTS_NAME],
            dim=dim,
            bin_dim_suffix="_bin",
        )

        # Add some coordinates to simplify interpretation/post-processing
        table = table.assign_coords(
            {
                OBSERVATIONS_NAME
                + "_bin": _get_category_bounds(self.observation_category_edges)
            }
        ).rename({OBSERVATIONS_NAME + "_bin": OBSERVATIONS_NAME + "_category_bounds"})
        table = table.assign_coords(
            {
                FORECASTS_NAME
                + "_bin": _get_category_bounds(self.forecast_category_edges)
            }
        ).rename({FORECASTS_NAME + "_bin": FORECASTS_NAME + "_category_bounds"})
        table = table.assign_coords(
            {
                OBSERVATIONS_NAME
                + "_category": (
                    OBSERVATIONS_NAME + "_category_bounds",
                    range(1, len(self.observation_category_edges)),
                ),
                FORECASTS_NAME
                + "_category": (
                    FORECASTS_NAME + "_category_bounds",
                    range(1, len(self.forecast_category_edges)),
                ),
            }
        )
        table = table.swap_dims(
            {
                OBSERVATIONS_NAME + "_category_bounds": OBSERVATIONS_NAME + "_category",
                FORECASTS_NAME + "_category_bounds": FORECASTS_NAME + "_category",
            }
        )

        return table

    def _sum_categories(self, categories):
        """Returns sums of specified categories in contingency table

        Parameters
        ----------
        category : str, optional
            Contingency table categories to sum.
            Options are 'total', 'observations' and 'forecasts'

        Returns
        -------
        Sum of all counts in specified categories

        """

        if categories == "total":
            N = self.table.sum(
                dim=(OBSERVATIONS_NAME + "_category", FORECASTS_NAME + "_category"),
                skipna=True,
            )
        elif categories == "observations":
            N = self.table.sum(dim=FORECASTS_NAME + "_category", skipna=True).rename(
                {OBSERVATIONS_NAME + "_category": "category"}
            )
        elif categories == "forecasts":
            N = self.table.sum(dim=OBSERVATIONS_NAME + "_category", skipna=True).rename(
                {FORECASTS_NAME + "_category": "category"}
            )
        else:
            raise ValueError(
                f"'{categories}' is not a recognised category. \
                    Pick one of ['total', 'observations', 'forecasts']"
            )

        return N

    def __repr__(self):
        return _display_metadata(self)

    @dichotomous_only
    def hits(self, yes_category=2):
        """Returns the number of hits (true positives) for dichotomous
        contingency data.

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the number of hits

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return self.table.sel(
            {
                OBSERVATIONS_NAME + "_category": yes_category,
                FORECASTS_NAME + "_category": yes_category,
            },
            drop=True,
        )

    @dichotomous_only
    def misses(self, yes_category=2):
        """Returns the number of misses (false negatives) for dichotomous
        contingency data.

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the number of misses

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """
        no_category = abs(yes_category - 2) + 1

        return self.table.sel(
            {
                OBSERVATIONS_NAME + "_category": yes_category,
                FORECASTS_NAME + "_category": no_category,
            },
            drop=True,
        )

    @dichotomous_only
    def false_alarms(self, yes_category=2):
        """Returns the number of false alarms (false positives) for dichotomous
        contingency data.

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the number of false alarms

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """
        no_category = abs(yes_category - 2) + 1

        return self.table.sel(
            {
                OBSERVATIONS_NAME + "_category": no_category,
                FORECASTS_NAME + "_category": yes_category,
            },
            drop=True,
        )

    @dichotomous_only
    def correct_negatives(self, yes_category=2):
        """Returns the number of correct negatives (true negatives) for dichotomous
        contingency data.

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the number of correct negatives

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """
        no_category = abs(yes_category - 2) + 1

        return self.table.sel(
            {
                OBSERVATIONS_NAME + "_category": no_category,
                FORECASTS_NAME + "_category": no_category,
            },
            drop=True,
        )

    @dichotomous_only
    def bias_score(self, yes_category=2):
        """Returns the bias score(s) for dichotomous contingency data

        .. math::

            BS = \\frac{\\mathrm{hits} + \\mathrm{false~alarms}}
                 {\\mathrm{hits} + \\mathrm{misses}}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the bias score(s)

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return (self.hits(yes_category) + self.false_alarms(yes_category)) / (
            self.hits(yes_category) + self.misses(yes_category)
        )

    @dichotomous_only
    def hit_rate(self, yes_category=2):
        """Returns the hit rate(s) (probability of detection) for
        dichotomous contingency data.

        .. math::
            HR = \\frac{hits}{hits + misses}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the
            category corresponding to 'yes' (1 or 2)

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the hit rate(s)

        See Also
        --------
        sklearn.metrics.recall_score

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return self.hits(yes_category) / (
            self.hits(yes_category) + self.misses(yes_category)
        )

    @dichotomous_only
    def false_alarm_ratio(self, yes_category=2):
        """Returns the false alarm ratio(s) for dichotomous contingency data.

        .. math::
            FAR = \\frac{\\mathrm{false~alarms}}{hits + \\mathrm{false~alarms}}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the false alarm ratio(s)

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return self.false_alarms(yes_category) / (
            self.hits(yes_category) + self.false_alarms(yes_category)
        )

    @dichotomous_only
    def false_alarm_rate(self, yes_category=2):
        """Returns the false alarm rate(s) (probability of false detection)
        for dichotomous contingency data.

        .. math::
            FA = \\frac{\\mathrm{false~alarms}}
                 {\\mathrm{correct~negatives} + \\mathrm{false~alarms}}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the false alarm rate(s)

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return self.false_alarms(yes_category) / (
            self.correct_negatives(yes_category) + self.false_alarms(yes_category)
        )

    @dichotomous_only
    def success_ratio(self, yes_category=2):
        """Returns the success ratio(s) for dichotomous contingency data.

        .. math::
            SR = \\frac{hits}{hits + \\mathrm{false~alarms}}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the success ratio(s)

        See Also
        --------
        sklearn.metrics.precision_score

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return self.hits(yes_category) / (
            self.hits(yes_category) + self.false_alarms(yes_category)
        )

    @dichotomous_only
    def threat_score(self, yes_category=2):
        """Returns the threat score(s) for dichotomous contingency data.

        .. math::
            TS = \\frac{hits}{hits + misses + \\mathrm{false~alarms}}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the threat score(s)

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return self.hits(yes_category) / (
            self.hits(yes_category)
            + self.misses(yes_category)
            + self.false_alarms(yes_category)
        )

    @dichotomous_only
    def equit_threat_score(self, yes_category=2):
        """Returns the equitable threat score(s) for dichotomous contingency data.

        .. math::
            ETS = \\frac{hits - hits_{random}}
                  {hits + misses + \\mathrm{false~alarms} - hits_{random}}

        .. math::
            hits_{random} = \\frac{(hits + misses
                            (hits + \\mathrm{false~alarms})}{total}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the equitable threat score(s)

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        hits_random = (
            (self.hits(yes_category) + self.misses(yes_category))
            * (self.hits(yes_category) + self.false_alarms(yes_category))
        ) / self._sum_categories("total")

        return (self.hits(yes_category) - hits_random) / (
            self.hits(yes_category)
            + self.misses(yes_category)
            + self.false_alarms(yes_category)
            - hits_random
        )

    @dichotomous_only
    def odds_ratio(self, yes_category=2):
        """Returns the odds ratio(s) for dichotomous contingency data

        .. math::
            OR = \\frac{hits * \\mathrm{correct~negatives}}
                       {misses * \\mathrm{false~alarms}}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the equitable odds ratio(s)

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return (self.hits(yes_category) * self.correct_negatives(yes_category)) / (
            self.misses(yes_category) * self.false_alarms(yes_category)
        )

    @dichotomous_only
    def odds_ratio_skill_score(self, yes_category=2):
        """Returns the odds ratio skill score(s) for dichotomous contingency data

        .. math::
            ORSS = \\frac{hits * \\mathrm{correct~negatives}
                      - misses * \\mathrm{false~alarms}}
            {hits * \\mathrm{correct~negatives} + misses * \\mathrm{false~alarms}}

        Parameters
        ----------
        yes_category : value, optional
            The category coordinate value of the category corresponding to 'yes'

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the equitable odds ratio skill score(s)

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        return (
            self.hits(yes_category) * self.correct_negatives(yes_category)
            - self.misses(yes_category) * self.false_alarms(yes_category)
        ) / (
            self.hits(yes_category) * self.correct_negatives(yes_category)
            + self.misses(yes_category) * self.false_alarms(yes_category)
        )

    def accuracy(self):
        """Returns the accuracy score(s) for a contingency table with K categories

        .. math::
            A = \\frac{1}{N}\\sum_{i=1}^{K} n(F_i, O_i)

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the accuracy score(s)

        See Also
        --------
        sklearn.metrics.accuracy_score

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        corr = self.table.where(
            self.table[OBSERVATIONS_NAME + "_category"]
            == self.table[FORECASTS_NAME + "_category"]
        ).sum(
            dim=(OBSERVATIONS_NAME + "_category", FORECASTS_NAME + "_category"),
            skipna=True,
        )
        N = self._sum_categories("total")

        return corr / N

    def heidke_score(self):
        """Returns the Heidke skill score(s) for a contingency table with K categories

        .. math::
            HSS = \\frac{\\frac{1}{N}\\sum_{i=1}^{K}n(F_i, O_i) -
                  \\frac{1}{N^2}\\sum_{i=1}^{K}N(F_i)N(O_i)}
                  {1 - \\frac{1}{N^2}\\sum_{i=1}^{K}N(F_i)N(O_i)}

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the Heidke score(s)

        See Also
        --------
        sklearn.metrics.cohen_kappa_score

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        numer_1 = (
            self.table.where(
                self.table[OBSERVATIONS_NAME + "_category"]
                == self.table[FORECASTS_NAME + "_category"]
            ).sum(
                dim=(OBSERVATIONS_NAME + "_category", FORECASTS_NAME + "_category"),
                skipna=True,
            )
            / self._sum_categories("total")
        )
        numer_2 = (
            self._sum_categories("observations") * self._sum_categories("forecasts")
        ).sum(dim="category", skipna=True) / self._sum_categories("total") ** 2
        denom = 1 - numer_2

        return (numer_1 - numer_2) / denom

    def peirce_score(self):
        """Returns the Peirce skill score(s) (Hanssen and Kuipers discriminantor true
        skill statistic) for a contingency table with K categories.

        .. math::
            PS = \\frac{\\frac{1}{N}\\sum_{i=1}^{K}n(F_i, O_i) -
                 \\frac{1}{N^2}\\sum_{i=1}^{K}N(F_i)N(O_i)}{1 -
                 \\frac{1}{N^2}\\sum_{i=1}^{K}N(O_i)^2}

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the Peirce score(s)

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """

        numer_1 = (
            self.table.where(
                self.table[OBSERVATIONS_NAME + "_category"]
                == self.table[FORECASTS_NAME + "_category"]
            ).sum(
                dim=(OBSERVATIONS_NAME + "_category", FORECASTS_NAME + "_category"),
                skipna=True,
            )
            / self._sum_categories("total")
        )
        numer_2 = (
            self._sum_categories("observations") * self._sum_categories("forecasts")
        ).sum(dim="category", skipna=True) / self._sum_categories("total") ** 2
        denom = 1 - (self._sum_categories("observations") ** 2).sum(
            dim="category", skipna=True
        ) / (self._sum_categories("total") ** 2)

        return (numer_1 - numer_2) / denom

    def gerrity_score(self):
        """Returns Gerrity equitable score for a contingency table with K categories.

        .. math::
                GS = \\frac{1}{N}\\sum_{i=1}^{K}\\sum_{j=1}^{K}n(F_i, O_j)s_{ij}

        .. math::
                s_{ii} = \\frac{1}{K-1}(\\sum_{r=1}^{i-1}a_r^{-1} +
                \\sum_{r=i}^{K-1}a_r)

        .. math::
                s_{ij} = \\frac{1}{K-1}(\\sum_{r=1}^{i-1}a_r^{-1} - (j - i) +
                \\sum_{r=j}^{K-1}a_r); 1 \\leq i < j \\leq K

        .. math::
            s_{ji} = s_{ij}

        .. math::
            a_i = \\frac{(1 - \\sum_{r=1}^{i}p_r)}{\\sum_{r=1}^{i}p_r}

        .. math::
            p_i = \\frac{N(O_i)}{N}

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            An array containing the Gerrity scores

        References
        ----------
        https://www.cawcr.gov.au/projects/verification/#Contingency_table
        """
        # TODO: Currently computes the Gerrity scoring matrix using nested for-loops.
        # Is it possible to remove these?

        def _gerrity_s(table):
            """Returns Gerrity scoring matrix, s"""
            p_o = (table.sum(axis=-1).T / table.sum(axis=(-2, -1)).T).T
            p_sum = np.cumsum(p_o, axis=-1)
            a = (1.0 - p_sum) / p_sum
            k = a.shape[-1]
            s = np.zeros(table.shape, dtype=float)
            for (i, j) in np.ndindex(*s.shape[-2:]):
                if i == j:
                    s[..., i, j] = (
                        1.0
                        / (k - 1.0)
                        * (
                            np.sum(1.0 / a[..., 0:j], axis=-1)
                            + np.sum(a[..., j : k - 1], axis=-1)
                        )
                    )
                elif i < j:
                    s[..., i, j] = (
                        1.0
                        / (k - 1.0)
                        * (
                            np.sum(1.0 / a[..., 0:i], axis=-1)
                            - (j - i)
                            + np.sum(a[..., j : k - 1], axis=-1)
                        )
                    )
                else:
                    s[..., i, j] = s[..., j, i]
            return s

        s = xr.apply_ufunc(
            _gerrity_s,
            self.table,
            input_core_dims=[
                [OBSERVATIONS_NAME + "_category", FORECASTS_NAME + "_category"]
            ],
            output_core_dims=[
                [OBSERVATIONS_NAME + "_category", FORECASTS_NAME + "_category"]
            ],
            dask="allowed",
        )

        return (self.table * s).sum(
            dim=(OBSERVATIONS_NAME + "_category", FORECASTS_NAME + "_category"),
            skipna=True,
        ) / self._sum_categories("total")



import bottleneck as bn
import numpy as np
from scipy import special
from scipy.stats import distributions

from .utils import suppress_warnings

__all__ = [
    "_effective_sample_size",
    "_linslope",
    "_mae",
    "_mape",
    "_me",
    "_median_absolute_error",
    "_mse",
    "_pearson_r",
    "_pearson_r_eff_p_value",
    "_pearson_r_p_value",
    "_r2",
    "_rmse",
    "_smape",
    "_spearman_r",
    "_spearman_r_eff_p_value",
    "_spearman_r_p_value",
]


def _match_nans(a, b, weights):
    """
    Considers missing values pairwise. If a value is missing
    in a, the corresponding value in b is turned to nan, and
    vice versa.

    Returns
    -------
    a, b, weights : ndarray
        a, b, and weights (if not None) with nans placed at
        pairwise locations.

    """
    if np.isnan(a).any() or np.isnan(b).any():
        # Avoids mutating original arrays and bypasses read-only issue.
        a, b = a.copy(), b.copy()
        # Find pairwise indices in a and b that have nans.
        idx = np.logical_or(np.isnan(a), np.isnan(b))
        a[idx], b[idx] = np.nan, np.nan
        # https://github.com/xarray-contrib/xskillscore/issues/168
        if isinstance(weights, np.ndarray):
            if weights.shape:  # not None
                weights = weights.copy()
                weights[idx] = np.nan
    return a, b, weights


def _get_numpy_funcs(skipna):
    """
    Returns nansum and nanmean if skipna is True;
    Returns sum and mean if skipna is False.
    """
    if skipna:
        return np.nansum, np.nanmean
    else:
        return np.sum, np.mean


def _check_weights(weights):
    """
    Quick check if weights are all NaN. If so,
    return None to guide weighting scheme.
    """
    if weights is None:
        return None
    # catch if np.ndarray values are None
    elif (weights == None).all():
        return None
    elif np.all(np.isnan(weights)):
        return None
    else:
        if weights.min() < 0:
            raise ValueError(
                "Weights has a minimum below 0. Please submit a weights array "
                "of positive numbers."
            )
        return weights


def __compute_anomalies(a, b, weights, axis, skipna):
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    # Only do weighted sums if there are weights. Cannot have a
    # single generic function with weights of all ones, because
    # the denominator gets inflated when there are masked regions.
    if weights is not None:
        with suppress_warnings("invalid value encountered in true_divide"):
            ma = sumfunc(a * weights, axis=axis) / sumfunc(weights, axis=axis)
            mb = sumfunc(b * weights, axis=axis) / sumfunc(weights, axis=axis)
    else:
        with suppress_warnings("Mean of empty slice"):
            ma = meanfunc(a, axis=axis)
            mb = meanfunc(b, axis=axis)
    am, bm = a - ma, b - mb
    return am, bm


def _effective_sample_size(a, b, axis, skipna):
    """Effective sample size for temporally correlated data.

    .. note::
        This metric should only be applied over the time dimension,
        since it is designed for temporal autocorrelation. Weights
        are not included due to the reliance on temporal
        autocorrelation.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to compute the effective sample size over.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    n_eff : ndarray
        Effective sample size.

    References
    ----------
    Bretherton, Christopher S., et al. "The effective number of spatial degrees of
      freedom of a time-varying field." Journal of climate 12.7 (1999): 1990-2009.
    Wilks, Daniel S. Statistical methods in the atmospheric sciences. Vol. 100.
      Academic press, 2011.
    """
    if skipna:
        a, b, _ = _match_nans(a, b, None)
    a = np.rollaxis(a, axis)
    b = np.rollaxis(b, axis)

    # count total number of samples that are non-nan.
    n = np.count_nonzero(~np.isnan(a), axis=0)

    # compute lag-1 autocorrelation.
    am, bm = __compute_anomalies(a, b, weights=None, axis=0, skipna=skipna)
    a_auto = _pearson_r(am[0:-1], am[1::], weights=None, axis=0, skipna=skipna)
    b_auto = _pearson_r(bm[0:-1], bm[1::], weights=None, axis=0, skipna=skipna)

    # compute effective sample size per Bretherton et al. 1999
    with suppress_warnings("divide by zero encountered in true_divide"):
        n_eff = n * (1 - a_auto * b_auto) / (1 + a_auto * b_auto)
    n_eff = np.floor(n_eff)
    n_eff = np.clip(n_eff, 0, n)
    return n_eff


def _linslope(a, b, weights, axis, skipna):
    """ndarray implementation of scipy.stats.linregress[slope].

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the linear slope along.
    weights : ndarray
        Input array of weights for a and b.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        slope of linear fit.

    See Also
    --------
    scipy.stats.linregress
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)
    a = np.rollaxis(a, axis)
    b = np.rollaxis(b, axis)
    if weights is not None:
        weights = np.rollaxis(weights, axis)

    am, bm = __compute_anomalies(a, b, weights=weights, axis=0, skipna=skipna)

    if weights is not None:
        s_num = sumfunc(weights * am * bm, axis=0)
        s_den = sumfunc(weights * am * am, axis=0)
    else:
        s_num = sumfunc(am * bm, axis=0)
        s_den = sumfunc(am * am, axis=0)

    with suppress_warnings("invalid value encountered in true_divide"):
        with suppress_warnings("invalid value encountered in double_scalars"):
            res = s_num / s_den
    return res


def _r2(a, b, weights, axis, skipna):
    """ndarray implementation of sklearn.metrics.r2_score.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the r2_score along.
    weights : ndarray
        Input array of weights for a and b.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        R^2 (coefficient of determination) score.

    See Also
    --------
    sklearn.metrics.r2_score

    References
    ----------
    https://en.wikipedia.org/wiki/Coefficient_of_determination
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)
    a = np.rollaxis(a, axis)
    b = np.rollaxis(b, axis)
    if weights is not None:
        weights = np.rollaxis(weights, axis)

    am, bm = __compute_anomalies(a, b, weights=weights, axis=0, skipna=skipna)

    if weights is not None:
        squared_error = weights * ((a - b) ** 2)
        am_squared = weights * (am ** 2)
    else:
        squared_error = (a - b) ** 2
        am_squared = am ** 2
    num = sumfunc(squared_error, axis=0)
    den = sumfunc(am_squared, axis=0)
    with suppress_warnings("invalid value encountered in true_divide"):
        with suppress_warnings("divide by zero encountered in true_divide"):
            with suppress_warnings("divide by zero encountered in double_scalars"):
                with suppress_warnings("invalid value encountered in double_scalars"):
                    r2 = 1 - (num / den)
    return r2


def _pearson_r(a, b, weights, axis, skipna):
    """ndarray implementation of scipy.stats.pearsonr.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the correlation along.
    weights : ndarray
        Input array of weights for a and b.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Pearson's correlation coefficient.

    See Also
    --------
    scipy.stats.pearsonr
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)
    a = np.rollaxis(a, axis)
    b = np.rollaxis(b, axis)
    if weights is not None:
        weights = np.rollaxis(weights, axis)

    am, bm = __compute_anomalies(a, b, weights=weights, axis=0, skipna=skipna)

    if weights is not None:
        r_num = sumfunc(weights * am * bm, axis=0)
        r_den = np.sqrt(
            sumfunc(weights * am * am, axis=0) * sumfunc(weights * bm * bm, axis=0)
        )
    else:
        r_num = sumfunc(am * bm, axis=0)
        r_den = np.sqrt(sumfunc(am * am, axis=0) * sumfunc(bm * bm, axis=0))

    with suppress_warnings("invalid value encountered in true_divide"):
        with suppress_warnings("invalid value encountered in double_scalars"):
            r = r_num / r_den
    res = np.clip(r, -1.0, 1.0)
    return res


def _pearson_r_p_value(a, b, weights, axis, skipna):
    """ndarray implementation of scipy.stats.pearsonr.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the compute the p value over.
    weights : ndarray
        Input array of weights for a and b.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        2-tailed p-value.

    See Also
    --------
    scipy.stats.pearsonr
    """
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    r = _pearson_r(a, b, weights, axis, skipna)
    if np.isnan(r).all():
        return r
    else:
        # no nans or some nans
        a = np.rollaxis(a, axis)
        b = np.rollaxis(b, axis)
        # count non-nans
        dof = np.count_nonzero(~np.isnan(a), axis=0) - 2
        with suppress_warnings("invalid value encountered in true_divide"):
            t_squared = r ** 2 * (dof / ((1.0 - r) * (1.0 + r)))
            _x = dof / (dof + t_squared)
        _x = np.asarray(_x)
        _x = np.where(_x < 1.0, _x, 1.0)
        _a = 0.5 * dof
        _b = 0.5
        res = special.betainc(_a, _b, _x)
        # reset masked values to nan
        # raises  <__array_function__ internals>:5: DeprecationWarning: Calling nonzero
        # on 0d arrays is deprecated, as it behaves surprisingly. Use
        # `atleast_1d(cond).nonzero()` if the old behavior was intended. If the context
        # of this warning is of the form `arr[nonzero(cond)]`, just use `arr[cond]`.
        nan_locs = np.where(np.isnan(r))
        if len(nan_locs[0]) > 0:
            res[nan_locs] = np.nan
        return res


def _pearson_r_eff_p_value(a, b, axis, skipna):
    """Pearson r p value accounting for autocorrelation.

    .. note::
        This metric should only be applied over the time dimension,
        since it is designed for temporal autocorrelation. Weights
        are not included due to the reliance on temporal
        autocorrelation.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to compute the p value over.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        2-tailed p-value.

    References
    ----------
    Bretherton, Christopher S., et al. "The effective number of spatial degrees of
      freedom of a time-varying field." Journal of climate 12.7 (1999): 1990-2009.
    Wilks, Daniel S. Statistical methods in the atmospheric sciences. Vol. 100.
      Academic press, 2011.
    """
    if skipna:
        a, b, _ = _match_nans(a, b, None)
    r = _pearson_r(a, b, None, axis, skipna)
    if np.isnan(r).all():
        return r
    else:
        dof = _effective_sample_size(a, b, axis, skipna) - 2
        t_squared = r ** 2 * (dof / ((1.0 - r) * (1.0 + r)))
        with suppress_warnings("invalid value encountered in true_divide"):
            _x = dof / (dof + t_squared)
        _x = np.asarray(_x)
        _x = np.where(_x < 1.0, _x, 1.0)
        _a = 0.5 * dof
        _b = 0.5
        res = special.betainc(_a, _b, _x)
        # reset masked values to nan
        nan_locs = np.where(np.isnan(r))
        if len(nan_locs[0]) > 0:
            res[nan_locs] = np.nan
        return res


def _spearman_r(a, b, weights, axis, skipna):
    """ndarray implementation of scipy.stats.spearmanr.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the correlation along.
    weights : ndarray
        Input array.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Spearmanr's correlation coefficient.

    See Also
    --------
    scipy.stats.spearmanr
    """
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    _a = bn.nanrankdata(a, axis=axis)
    _b = bn.nanrankdata(b, axis=axis)
    return _pearson_r(_a, _b, weights, axis, skipna)


def _spearman_r_p_value(a, b, weights, axis, skipna):
    """ndarray implementation of scipy.stats.spearmanr.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the correlation along.
    weights : ndarray
        Input array.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        2-tailed p-value.

    See Also
    --------
    scipy.stats.spearmanr

    References
    ----------
    https://github.com/scipy/scipy/blob/v1.3.1/scipy/stats/stats.py#L3613-L3764
    """
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    rs = _spearman_r(a, b, weights, axis, skipna)
    a = np.rollaxis(a, axis)
    b = np.rollaxis(b, axis)
    # count non-nans
    dof = np.count_nonzero(~np.isnan(a), axis=0) - 2
    with suppress_warnings("invalid value encountered in true_divide"):
        with suppress_warnings("divide by zero encountered in true_divide"):
            t = rs * np.sqrt((dof / ((rs + 1.0) * (1.0 - rs))).clip(0))
    p = 2 * distributions.t.sf(np.abs(t), dof)
    return p


def _spearman_r_eff_p_value(a, b, axis, skipna):
    """Spearman rank correlation p value, accounting for autocorrelation.

    .. note::
        This metric should only be applied over the time dimension,
        since it is designed for temporal autocorrelation. Weights
        are not included due to the reliance on temporal
        autocorrelation.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    weights : ndarray
        Input array.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        2-tailed p-value.

    See Also
    --------
    scipy.stats.spearmanr

    References
    ----------
    Bretherton, Christopher S., et al. "The effective number of spatial degrees of
      freedom of a time-varying field." Journal of climate 12.7 (1999): 1990-2009.
    Wilks, Daniel S. Statistical methods in the atmospheric sciences. Vol. 100.
      Academic press, 2011.
    """
    if skipna:
        a, b, _ = _match_nans(a, b, None)
    rs = _spearman_r(a, b, None, axis, skipna)
    dof = _effective_sample_size(a, b, axis, skipna) - 2
    with suppress_warnings("invalid value encountered in true_divide"):
        with suppress_warnings("divide by zero encountered in true_divide"):
            t = rs * np.sqrt((dof / ((rs + 1.0) * (1.0 - rs))).clip(0))
    p = 2 * distributions.t.sf(np.abs(t), dof)
    return p


def _me(a, b, weights, axis, skipna):
    """Mean Error.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the me along.
    weights : ndarray
        Input array of weights for a and b.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Mean Error.
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)

    error = a - b
    if weights is not None:
        with suppress_warnings("invalid value encountered in true_divide"):
            mean_error = sumfunc(error * weights, axis=axis) / sumfunc(
                weights, axis=axis
            )
    else:
        with suppress_warnings("Mean of empty slice"):
            mean_error = meanfunc(error, axis=axis)
    return mean_error


def _rmse(a, b, weights, axis, skipna):
    """Root Mean Squared Error.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the rmse along.
    weights : ndarray
        Input array of weights for a and b.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Root Mean Squared Error.

    See Also
    --------
    sklearn.metrics.mean_squared_error
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)

    squared_error = (a - b) ** 2
    if weights is not None:
        with suppress_warnings("invalid value encountered in true_divide"):
            mean_squared_error = sumfunc(squared_error * weights, axis=axis) / sumfunc(
                weights, axis=axis
            )
    else:
        with suppress_warnings("Mean of empty slice"):
            mean_squared_error = meanfunc(squared_error, axis=axis)
    res = np.sqrt(mean_squared_error)
    return res


def _mse(a, b, weights, axis, skipna):
    """Mean Squared Error.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the mse along.
    weights : ndarray
        Input array of weights for a and b.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Mean Squared Error.

    See Also
    --------
    sklearn.metrics.mean_squared_error
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)

    squared_error = (a - b) ** 2
    if weights is not None:
        with suppress_warnings("invalid value encountered in true_divide"):
            return sumfunc(squared_error * weights, axis=axis) / sumfunc(
                weights, axis=axis
            )
    else:
        with suppress_warnings("Mean of empty slice"):
            return meanfunc(squared_error, axis=axis)


def _mae(a, b, weights, axis, skipna):
    """Mean Absolute Error.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the mae along.
    weights : ndarray
        Input array of weights for a and b.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Mean Absolute Error.

    See Also
    --------
    sklearn.metrics.mean_absolute_error
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)

    absolute_error = np.absolute(a - b)
    if weights is not None:
        with suppress_warnings("invalid value encountered in true_divide"):
            return sumfunc(absolute_error * weights, axis=axis) / sumfunc(
                weights, axis=axis
            )
    else:
        with suppress_warnings("Mean of empty slice"):
            return meanfunc(absolute_error, axis=axis)


def _median_absolute_error(a, b, axis, skipna):
    """Median Absolute Error.

    Parameters
    ----------
    a : ndarray
        Input array.
    b : ndarray
        Input array.
    axis : int
        The axis to apply the median absolute error along.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Median Absolute Error.

    See Also
    --------
    sklearn.metrics.median_absolute_error
    """
    medianfunc = np.nanmedian if skipna else np.median
    if skipna:
        a, b, _ = _match_nans(a, b, None)
    absolute_error = np.absolute(a - b)
    with suppress_warnings("All-NaN slice encountered"):
        return medianfunc(absolute_error, axis=axis)


def _mape(a, b, weights, axis, skipna):
    """Mean Absolute Percentage Error.

    .. math::
        \\mathrm{MAPE} = \\frac{1}{n} \\sum_{i=1}^{n}
                         \\frac{\\vert a_{i} - b_{i} \\vert}
                               {max(\epsilon, \\vert a_{i} \\vert)}

    Parameters
    ----------
    a : ndarray
        Input array (truth to be divided by).
    b : ndarray
        Input array.
    axis : int
        The axis to apply the mape along.
    weights : ndarray
        Input array.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Mean Absolute Percentage Error.

    Notes
    -----
    The percent error is calculated in reference to ``a``.

    Percent error is reported as decimal percent. I.e., a value of
    1 is 100%.

    \epsilon is an arbitrary small yet strictly positive number to avoid
    undefined results when ``a`` is zero.

    See Also
    --------
    sklearn.metrics.mean_absolute_percentage_error

    References
    ----------
    https://en.wikipedia.org/wiki/Mean_absolute_percentage_error
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)
    epsilon = np.finfo(np.float64).eps
    mape = np.absolute(a - b) / np.maximum(np.absolute(a), epsilon)
    if weights is not None:
        with suppress_warnings("invalid value encountered in true_divide"):
            return sumfunc(mape * weights, axis=axis) / sumfunc(weights, axis=axis)
    else:
        with suppress_warnings("Mean of empty slice"):
            return meanfunc(mape, axis=axis)


def _smape(a, b, weights, axis, skipna):
    """Symmetric Mean Absolute Percentage Error.

    .. math::
        \\mathrm{SMAPE} = \\frac{1}{n} \\sum_{i=1}^{n}
                          \\frac{ \\vert a_{i} - b_{i} \\vert }
                          { \\vert a_{i} \\vert + \\vert b_{i} \\vert  }

    Parameters
    ----------
    a : ndarray
        Input array (truth to be divided by).
    b : ndarray
        Input array.
    axis : int
        The axis to apply the mae along.
    weights : ndarray
        Input array.
    skipna : bool
        If True, skip NaNs when computing function.

    Returns
    -------
    res : ndarray
        Symmetric Mean Absolute Percentage Error.

    Notes
    -----
    Symmetric percent error is reported as decimal percent. I.e., a value of 1
    is 100%.

    References
    ----------
    https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error
    """
    sumfunc, meanfunc = _get_numpy_funcs(skipna)
    if skipna:
        a, b, weights = _match_nans(a, b, weights)
    weights = _check_weights(weights)
    smape = np.absolute(a - b) / (np.absolute(a) + np.absolute(b))
    if weights is not None:
        with suppress_warnings("invalid value encountered in true_divide"):
            return sumfunc(smape * weights, axis=axis) / sumfunc(weights, axis=axis)
    else:
        with suppress_warnings("Mean of empty slice"):
            return meanfunc(smape, axis=axis)



from typing import Callable, List

import dask
import numpy as np
import pytest
from xarray.tests import CountingScheduler, assert_allclose, raise_if_dask_computes

from xskillscore.core.deterministic import (
    linslope,
    mae,
    mape,
    me,
    median_absolute_error,
    mse,
    pearson_r,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_p_value,
)

WEIGHTED_METRICS: List[Callable] = [
    linslope,
    pearson_r,
    pearson_r_p_value,
    spearman_r,
    spearman_r_p_value,
    mae,
    mse,
    mape,
    smape,
    me,
    rmse,
    r2,
]

NON_WEIGHTED_METRICS: List[Callable] = [median_absolute_error]


def drop_nans(a, b, weights=None, dim="time"):
    """
    Masks a and b where they have pairwise nans.
    """
    a = a.where(b.notnull())
    b = b.where(a.notnull())
    if weights is not None:
        weights = weights.where(a.notnull())
        weights = weights.dropna(dim)
    return a.dropna(dim), b.dropna(dim), weights


@pytest.mark.parametrize("metric", WEIGHTED_METRICS + NON_WEIGHTED_METRICS)
def test_skipna_returns_same_value_as_dropped_pairwise_nans(
    a_1d_fixed_nan, b_1d_fixed_nan, metric
):
    """Tests that DataArrays with pairwise nans return the same result
    as the same two with those nans dropped."""
    a_dropped, b_dropped, _ = drop_nans(a_1d_fixed_nan, b_1d_fixed_nan)
    with raise_if_dask_computes():
        res_with_nans = metric(a_1d_fixed_nan, b_1d_fixed_nan, "time", skipna=True)
        res_dropped_nans = metric(a_dropped, b_dropped, "time")
    assert_allclose(res_with_nans, res_dropped_nans)


@pytest.mark.parametrize("metric", WEIGHTED_METRICS)
def test_skipna_returns_same_value_as_dropped_pairwise_nans_with_weights(
    a_1d_fixed_nan, b_1d_fixed_nan, weights_time, metric
):
    """Tests that DataArrays with pairwise nans return the same result
    as the same two with those nans dropped."""
    a_dropped, b_dropped, weights_time_dropped = drop_nans(
        a_1d_fixed_nan, b_1d_fixed_nan, weights_time
    )
    with raise_if_dask_computes():
        res_with_nans = metric(
            a_1d_fixed_nan, b_1d_fixed_nan, "time", skipna=True, weights=weights_time
        )
        res_dropped_nans = metric(
            a_dropped, b_dropped, "time", weights=weights_time_dropped
        )
    assert_allclose(res_with_nans, res_dropped_nans)


@pytest.mark.parametrize("metric", WEIGHTED_METRICS + NON_WEIGHTED_METRICS)
def test_skipna_returns_nan_when_false(a_1d_fixed_nan, b_1d_fixed_nan, metric):
    """Tests that nan is returned if there's any nans in the time series
    and skipna is False."""
    with raise_if_dask_computes():
        res = metric(a_1d_fixed_nan, b_1d_fixed_nan, "time", skipna=False)
    assert np.isnan(res).all()


@pytest.mark.parametrize("metric", WEIGHTED_METRICS)
def test_skipna_broadcast_weights_assignment_destination(
    a_rand_nan, b_rand_nan, weights_lonlat, metric
):
    """Tests that 'assignment destination is read-only' is not raised
    https://github.com/xarray-contrib/xskillscore/issues/79"""
    with raise_if_dask_computes():
        metric(
            a_rand_nan, b_rand_nan, ["lat", "lon"], weights=weights_lonlat, skipna=True
        )


def test_nan_skipna(a, b):
    # Randomly add some nans to a
    a = a.where(np.random.random(a.shape) < 0.5)
    with raise_if_dask_computes():
        pearson_r(a, b, dim="lat", skipna=True)



import pytest
import xarray as xr
from xarray.tests import assert_allclose

from xskillscore.core.deterministic import (
    effective_sample_size,
    linslope,
    mae,
    mape,
    me,
    median_absolute_error,
    mse,
    pearson_r,
    pearson_r_eff_p_value,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_eff_p_value,
    spearman_r_p_value,
)

correlation_metrics = [
    linslope,
    pearson_r,
    r2,
    pearson_r_p_value,
    spearman_r,
    spearman_r_p_value,
    effective_sample_size,
    pearson_r_eff_p_value,
    spearman_r_eff_p_value,
]

temporal_only_metrics = [
    pearson_r_eff_p_value,
    spearman_r_eff_p_value,
    effective_sample_size,
]

distance_metrics = [
    me,
    mse,
    rmse,
    mae,
    median_absolute_error,
    mape,
    smape,
]

AXES = ("time", "lat", "lon", ["lat", "lon"], ["time", "lat", "lon"])


def _ds(a, b, skipna_bool):
    ds = xr.Dataset()
    ds["a"] = a
    ds["b"] = b
    if skipna_bool is True:
        ds["b"] = b.where(b < 0.5)
    return ds


def adjust_weights(dim, weight_bool, weights):
    """
    Adjust the weights test data to only span the core dimension
    that the function is being applied over.
    """
    if weight_bool:
        drop_dims = [i for i in weights.dims if i not in dim]
        drop_dims = {k: 0 for k in drop_dims}
        return weights.isel(drop_dims)
    else:
        return None


@pytest.mark.parametrize("outer_bool", [False, True])
@pytest.mark.parametrize("metric", correlation_metrics + distance_metrics)
@pytest.mark.parametrize("dim", AXES)
@pytest.mark.parametrize("weight_bool", [False, True])
@pytest.mark.parametrize("skipna_bool", [False, True])
def test_deterministic_metrics_accessor(
    a, b, dim, skipna_bool, weight_bool, weights_cos_lat, metric, outer_bool
):

    # Update dim to time if testing temporal only metrics
    if (dim != "time") and (metric in temporal_only_metrics):
        dim = "time"

    _weights = adjust_weights(dim, weight_bool, weights_cos_lat)
    ds = _ds(a, b, skipna_bool)
    b = ds["b"]  # Update if populated with nans
    if outer_bool:
        ds = ds.drop_vars("b")

    accessor_func = getattr(ds.xs, metric.__name__)
    if metric in temporal_only_metrics or metric == median_absolute_error:
        actual = metric(a, b, dim=dim, skipna=skipna_bool)
        if outer_bool:
            expected = accessor_func("a", b, dim=dim, skipna=skipna_bool)
        else:
            expected = accessor_func("a", "b", dim=dim, skipna=skipna_bool)
    else:
        actual = metric(a, b, dim=dim, weights=_weights, skipna=skipna_bool)
        if outer_bool:
            expected = accessor_func(
                "a", b, dim=dim, weights=_weights, skipna=skipna_bool
            )
        else:
            expected = accessor_func(
                "a", "b", dim=dim, weights=_weights, skipna=skipna_bool
            )
    assert_allclose(actual, expected)



# https://github.com/pydata/xarray/blob/master/asv_bench/benchmarks/__init__.py

import numpy as np
import xarray as xr


def parameterized(names, params):
    def decorator(func):
        func.param_names = names
        func.params = params
        return func

    return decorator


def requires_dask():
    try:
        import dask  # noqa
    except ImportError:
        raise NotImplementedError


def randn(shape, frac_nan=None, chunks=None, seed=0):
    rng = np.random.RandomState(seed)
    if chunks is None:
        x = rng.standard_normal(shape)
    else:
        import dask.array as da

        rng = da.random.RandomState(seed)
        x = rng.standard_normal(shape, chunks=chunks)

    if frac_nan is not None:
        inds = rng.choice(range(x.size), int(x.size * frac_nan))
        x.flat[inds] = np.nan

    return x


def randint(low, high=None, size=None, frac_minus=None, seed=0):
    rng = np.random.RandomState(seed)
    x = rng.randint(low, high, size)
    if frac_minus is not None:
        inds = rng.choice(range(x.size), int(x.size * frac_minus))
        x.flat[inds] = -1

    return x


class Generate:
    """
    Generate random xr.Dataset ds to be benckmarked.
    """

    timeout = 600
    repeat = (2, 5, 20)

    def make_ds(self, nmember, nx, ny, chunks=None):

        # ds
        self.ds = xr.Dataset()
        self.nmember = nmember
        self.nx = nx
        self.ny = ny

        frac_nan = 0.0

        members = np.arange(1, 1 + self.nmember)

        lons = xr.DataArray(
            np.linspace(0, 360, self.nx),
            dims=("lon",),
            attrs={"units": "degrees east", "long_name": "longitude"},
        )
        lats = xr.DataArray(
            np.linspace(-90, 90, self.ny),
            dims=("lat",),
            attrs={"units": "degrees north", "long_name": "latitude"},
        )
        self.ds["tos"] = xr.DataArray(
            randn((self.nmember, self.nx, self.ny), frac_nan=frac_nan, chunks=chunks),
            coords={"member": members, "lon": lons, "lat": lats},
            dims=("member", "lon", "lat"),
            name="tos",
            attrs={"units": "foo units", "description": "a description"},
        )
        self.ds["sos"] = xr.DataArray(
            randn((self.nmember, self.nx, self.ny), frac_nan=frac_nan, chunks=chunks),
            coords={"member": members, "lon": lons, "lat": lats},
            dims=("member", "lon", "lat"),
            name="sos",
            attrs={"units": "foo units", "description": "a description"},
        )
        self.ds.attrs = {"history": "created for xskillscore benchmarking"}

        # set nans for land sea mask
        self.ds = self.ds.where(
            (abs(self.ds.lat) > 20) | (self.ds.lat < 100) | (self.ds.lat > 160)
        )



import numpy as np
import pytest
from sklearn.metrics import (
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    r2_score,
)

import xskillscore as xs
from xskillscore.core.deterministic import (
    linslope,
    mae,
    mape,
    me,
    mse,
    pearson_r,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_p_value,
)

xs_skl_metrics = [
    (r2, r2_score),
    (mse, mean_squared_error),
    (mae, mean_absolute_error),
    (mape, mean_absolute_percentage_error),
]

xs_skl_metrics_with_zeros = [
    (mape, mean_absolute_percentage_error),
]


def weighted_pearsonr(x, y, w):
    xm = x - (np.sum(x * w) / np.sum(w))
    ym = y - (np.sum(y * w) / np.sum(w))
    r_num = np.sum(w * xm * ym)
    r_den = np.sqrt(np.sum(w * xm * xm) * np.sum(w * ym * ym))
    return r_num / r_den


def weighted_linslope(x, y, w):
    xm = x - (np.sum(x * w) / np.sum(w))
    ym = y - (np.sum(y * w) / np.sum(w))
    s_num = np.sum(w * xm * ym)
    s_den = np.sum(w * xm * xm)
    return s_num / s_den


xs_scipy_metrics = [(pearson_r, weighted_pearsonr), (linslope, weighted_linslope)]


xs_np_metrics = [
    (me, lambda x, y, w: np.sum((x - y) * w) / np.sum(w)),
    (
        smape,
        lambda x, y, w: np.sum(np.abs(x - y) / (np.abs(x) + np.abs(y)) * w) / np.sum(w),
    ),
]


@pytest.mark.parametrize("xs_skl_metrics", xs_skl_metrics)
def test_xs_same_as_skl_weighted(a_1d, b_1d, weights_linear_time_1d, xs_skl_metrics):
    """Tests weighted xskillscore metric is same as weighted scikit-learn metric."""
    xs_metric, skl_metric = xs_skl_metrics
    actual = xs_metric(a_1d, b_1d, "time", weights_linear_time_1d)
    expected = skl_metric(a_1d, b_1d, sample_weight=weights_linear_time_1d)
    assert np.allclose(actual, expected)


def test_xs_same_as_skl_rmse_weighted(a_1d, b_1d, weights_linear_time_1d):
    actual = rmse(a_1d, b_1d, "time", weights_linear_time_1d)
    expected = mean_squared_error(
        a_1d, b_1d, squared=False, sample_weight=weights_linear_time_1d
    )
    assert np.allclose(actual, expected)


@pytest.mark.parametrize("xs_skl_metrics", xs_skl_metrics_with_zeros)
def test_xs_same_as_skl_with_zeros_weighted(
    a_1d_with_zeros, b_1d, xs_skl_metrics, weights_linear_time_1d
):
    """Tests weighted xskillscore metric is same as weighted scikit-learn metric."""
    xs_metric, skl_metric = xs_skl_metrics
    actual = xs_metric(a_1d_with_zeros, b_1d, "time", weights_linear_time_1d)
    expected = skl_metric(a_1d_with_zeros, b_1d, sample_weight=weights_linear_time_1d)
    assert np.allclose(actual, expected)


@pytest.mark.parametrize("xs_scipy_metrics", xs_scipy_metrics)
def test_xs_same_as_scipy(a_1d, b_1d, xs_scipy_metrics, weights_linear_time_1d):
    """Tests weighted xskillscore metric is same as weighted scipy metric."""
    xs_metric, scipy_metric = xs_scipy_metrics
    actual = xs_metric(a_1d, b_1d, "time", weights_linear_time_1d)
    expected = scipy_metric(a_1d.values, b_1d.values, weights_linear_time_1d.values)
    assert np.allclose(actual, expected)


@pytest.mark.parametrize("xs_np_metrics", xs_np_metrics)
def test_xs_same_as_numpy_weighted(a_1d, b_1d, xs_np_metrics, weights_linear_time_1d):
    """Tests weighted xskillscore metric is same as weighted metric using numpy."""
    xs_metric, np_metric = xs_np_metrics
    actual = xs_metric(a_1d, b_1d, "time", weights_linear_time_1d)
    expected = np_metric(a_1d, b_1d, weights_linear_time_1d)
    assert np.allclose(actual, expected)



import contextlib
import warnings

import numpy as np
import xarray as xr
from xhistogram.xarray import histogram as xhist

__all__ = ["histogram"]


@contextlib.contextmanager
def suppress_warnings(msg=None):
    """Catch warnings with message msg. From
    https://github.com/TheClimateCorporation/properscoring/blob/master/properscoring/
    _utils.py#L23."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", msg)
        yield


def _preprocess_dims(dim, a):
    """Preprocesses dimensions to prep for stacking.
    Parameters
    ----------
    dim : str, list
        The dimension(s) to apply the function along.
    """
    if dim is None:
        dim = list(a.dims)
    elif isinstance(dim, str):
        dim = [dim]
    axis = tuple(range(-1, -len(dim) - 1, -1))
    return dim, axis


def _fail_if_dim_empty(dim):
    if dim == []:
        raise ValueError(
            "metric must be applied along one dimension, therefore "
            f"requires `dim` not being empty, found dim={dim}"
        )


def _stack_input_if_needed(a, b, dim, weights):
    """
    Stack input arrays a, b if needed in correlation metrics.
    Adapt dim and weights accordingly.
    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : list
        The dimension(s) to apply the correlation along.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    Returns
    -------
    a : xarray.Dataset or xarray.DataArray stacked with new_dim
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray stacked with new_dim
        Labeled array(s) over which to apply the function.
    new_dim : str
        The dimension(s) to apply the correlation along.
    weights : xarray.Dataset or xarray.DataArray stacked with new_dim or None
        Weights matching dimensions of ``dim`` to apply during the function.
    """
    if len(dim) > 1:
        new_dim = "_".join(dim)
        a = a.stack(**{new_dim: dim})
        b = b.stack(**{new_dim: dim})
        if weights is not None:
            weights = weights.stack(**{new_dim: dim})
    else:
        new_dim = dim[0]
    return a, b, new_dim, weights


def _preprocess_weights(a, dim, new_dim, weights):
    """Preprocesses weights array to prepare for numpy computation.
    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        One of the arrays over which the function will be applied.
    dim : str, list
        The original dimension(s) to apply the function along.
    new_dim : str
        The newly named dimension after running ``_preprocess_dims``
    weights : xarray.Dataset or xarray.DataArray or None
        Weights to apply to function, matching the dimension size of
        ``new_dim``.
    """
    if weights is None:
        return None
    else:
        # Scale weights to vary from 0 to 1.
        weights = weights / weights.max()
        # Check that the weights array has the same size
        # dimension(s) as those being applied over.
        drop_dims = {k: 0 for k in a.dims if k not in new_dim}
        if dict(weights.sizes) != dict(a.isel(drop_dims).sizes):
            raise ValueError(
                f"weights dimension(s) {dim} of size {dict(weights.sizes)} "
                f"does not match DataArray's size "
                f"{dict(a.isel(drop_dims).sizes)}"
            )
        if dict(weights.sizes) != dict(a.sizes):
            # Broadcast weights to full size of main object.
            _, weights = xr.broadcast(a, weights)
        return weights


def _add_as_coord(ds1, ds2, coordinate_suffix):
    """Add ds2 as a coordinate of ds1.
    Assumes that ds1 and ds2 are the same type of xarray object
    """
    if isinstance(ds1, xr.Dataset):
        for var in ds1.data_vars:
            ds1 = ds1.assign_coords({f"{ds2[var].name}_{coordinate_suffix}": ds2[var]})
    elif isinstance(ds1, xr.DataArray):
        ds1 = ds1.assign_coords({coordinate_suffix: ds2})
    else:
        raise ValueError("Inputs ds1 and ds2 must be xarray objects")
    return ds1


def _get_bin_centers(bin_edges):
    """Return the arithmetic mean of the bin_edges"""
    return 0.5 * (bin_edges[:-1] + bin_edges[1:])


def histogram(*args, bins=None, bin_names=None, **kwargs):
    """Wrapper on xhistogram to deal with Datasets appropriately"""
    # xhistogram expects a list for the dim input
    if "dim" in kwargs:
        if isinstance(kwargs["dim"], str):
            kwargs["dim"] = [kwargs["dim"]]
    for bin in bins:
        assert isinstance(
            bin, np.ndarray
        ), f"all bins must be numpy arrays, found {type(bin)}"

    if isinstance(args[0], xr.Dataset):
        # Get list of variables that are shared across all Datasets
        overlapping_vars = set.intersection(*map(set, [arg.data_vars for arg in args]))
        if overlapping_vars:
            # If bin_names not provided, use default ----
            if bin_names is None:
                bin_names = ["ds_" + str(i + 1) for i in range(len(args))]
            return xr.merge(
                [
                    xhist(
                        *(arg[var].rename(bin_names[i]) for i, arg in enumerate(args)),
                        bins=bins,
                        **kwargs,
                    ).rename(var)
                    for var in overlapping_vars
                ]
            )
        else:
            raise ValueError("No common variables exist across input Datasets")
    else:
        if bin_names:
            args = (arg.rename(bin_names[i]) for i, arg in enumerate(args))
        return xhist(*args, bins=bins, **kwargs)


def _bool_to_int(ds):
    """convert xr.object of dtype bool to int to evade:
    TypeError: numpy boolean subtract, the `-` operator, is not supported"""

    def _helper_bool_to_int(da):
        if da.dtype == "bool":
            da = da.astype("int")
        return da

    if isinstance(ds, xr.Dataset):
        ds = ds.map(_helper_bool_to_int)
    else:
        ds = _helper_bool_to_int(ds)
    return ds


def _check_identical_xr_types(a, b):
    """Check that a and b are both xr.Dataset or both xr.DataArray."""
    if type(a) != type(b):
        raise ValueError(f"a and b must be same type, found {type(a)} and {type(b)}")
    for d in [a, b]:
        if not isinstance(d, (xr.Dataset, xr.DataArray)):
            raise ValueError("inputs must be xr.DataArray or xr.Dataset")


def _keep_nans_masked(ds_before, ds_after, dim=None, ignore=None):
    """Preserve all NaNs from ds_before for ds_after over while ignoring some dimensions optionally."""
    if dim is None:
        dim = list(ds_before.dims)
    elif isinstance(dim, str):
        dim = [dim]
    if ignore is None:
        ignore = []
    elif isinstance(ignore, str):
        ignore = list(ignore)
    all_dim = set(dim) ^ set(ignore)
    all_dim = [d for d in all_dim if d in ds_before.dims]
    mask = ds_before.isnull().all(all_dim)
    ds_after = ds_after.where(~mask.astype("bool"), other=np.nan)
    for d in dim:
        assert d not in ds_after.dims
    if ignore is not None:
        for d in ignore:
            assert d not in ds_after.dims
    return ds_after



import numpy as np
import pytest
import xarray as xr
from scipy.stats import norm
from xarray.tests import assert_allclose

from xskillscore.core.probabilistic import (
    brier_score,
    crps_ensemble,
    crps_gaussian,
    crps_quadrature,
    discrimination,
    rank_histogram,
    reliability,
    roc,
    rps,
    threshold_brier_score,
)


@pytest.fixture
def threshold():
    return 0.5


@pytest.mark.parametrize("outer_bool", [False, True])
def test_crps_gaussian_accessor(o, f_prob, outer_bool):
    mu = f_prob.mean("member")
    sig = f_prob.std("member")
    actual = crps_gaussian(o, mu, sig)
    ds = xr.Dataset()
    ds["o"] = o
    ds["mu"] = mu
    ds["sig"] = sig
    if outer_bool:
        ds = ds.drop_vars("mu")
        expected = ds.xs.crps_gaussian("o", mu, sig)
    else:
        expected = ds.xs.crps_gaussian("o", "mu", "sig")
    assert_allclose(actual, expected)


@pytest.mark.parametrize("outer_bool", [False, True])
def test_crps_ensemble_accessor(o, f_prob, outer_bool):
    actual = crps_ensemble(o, f_prob)
    ds = xr.Dataset()
    ds["o"] = o
    ds["f_prob"] = f_prob
    if outer_bool:
        ds = ds.drop_vars("f_prob")
        expected = ds.xs.crps_ensemble("o", f_prob)
    else:
        expected = ds.xs.crps_ensemble("o", "f_prob")
    assert_allclose(actual, expected)


@pytest.mark.slow
def test_crps_quadrature_accessor(o):
    # to speed things up
    o = o.isel(time=0, drop=True)
    cdf_or_dist = norm
    actual = crps_quadrature(o, cdf_or_dist)
    ds = xr.Dataset()
    ds["o"] = o
    expected = ds.xs.crps_quadrature("o", cdf_or_dist)
    assert_allclose(actual, expected)


@pytest.mark.parametrize("outer_bool", [False, True])
def test_threshold_brier_score_accessor(o, f_prob, threshold, outer_bool):
    actual = threshold_brier_score(o, f_prob, threshold)
    ds = xr.Dataset()
    ds["o"] = o
    ds["f_prob"] = f_prob
    if outer_bool:
        ds = ds.drop_vars("f_prob")
        expected = ds.xs.threshold_brier_score("o", f_prob, threshold)
    else:
        expected = ds.xs.threshold_brier_score("o", "f_prob", threshold)
    assert_allclose(actual, expected)


@pytest.mark.parametrize("outer_bool", [False, True])
def test_brier_score_accessor(o, f_prob, threshold, outer_bool):
    actual = brier_score(o > threshold, (f_prob > threshold).mean("member"))
    ds = xr.Dataset()
    ds["o"] = o > threshold
    ds["f_prob"] = (f_prob > threshold).mean("member")
    if outer_bool:
        ds = ds.drop_vars("f_prob")
        expected = ds.xs.brier_score("o", (f_prob > threshold).mean("member"))
    else:
        expected = ds.xs.brier_score("o", "f_prob")
    assert_allclose(actual, expected)


@pytest.mark.parametrize("outer_bool", [False, True])
def test_rank_histogram_accessor(o, f_prob, outer_bool):
    actual = rank_histogram(o, f_prob)
    ds = xr.Dataset()
    ds["o"] = o
    ds["f_prob"] = f_prob
    if outer_bool:
        ds = ds.drop_vars("f_prob")
        expected = ds.xs.rank_histogram("o", f_prob)
    else:
        expected = ds.xs.rank_histogram("o", "f_prob")
    assert_allclose(actual, expected)


@pytest.mark.parametrize("outer_bool", [False, True])
def test_discrimination_accessor(o, f_prob, threshold, outer_bool):
    actual = discrimination(o > threshold, (f_prob > threshold).mean("member"))
    ds = xr.Dataset()
    ds["o"] = o > threshold
    ds["f_prob"] = (f_prob > threshold).mean("member")
    if outer_bool:
        ds = ds.drop_vars("f_prob")
        expected = ds.xs.discrimination("o", (f_prob > threshold).mean("member"))
    else:

        expected = ds.xs.discrimination("o", "f_prob")
    assert_allclose(actual, expected)


@pytest.mark.parametrize("outer_bool", [False, True])
def test_reliability_accessor(o, f_prob, threshold, outer_bool):
    actual = reliability(o > threshold, (f_prob > threshold).mean("member"))

    ds = xr.Dataset()
    ds["o"] = o > threshold
    ds["f_prob"] = (f_prob > threshold).mean("member")
    if outer_bool:
        ds = ds.drop_vars("f_prob")
        expected = ds.xs.reliability("o", (f_prob > threshold).mean("member"))
    else:
        expected = ds.xs.reliability("o", "f_prob")
    assert_allclose(actual, expected)


@pytest.mark.parametrize("outer_bool", [False, True])
def test_rps_accessor(o, f_prob, outer_bool):
    category_edges = np.linspace(0, 1, 6)
    actual = rps(o, f_prob, category_edges=category_edges)

    ds = xr.Dataset()
    ds["o"] = o
    ds["f_prob"] = f_prob
    if outer_bool:
        ds = ds.drop_vars("f_prob")
        expected = ds.xs.rps("o", f_prob, category_edges=category_edges)
    else:
        expected = ds.xs.rps("o", "f_prob", category_edges=category_edges)
    assert_allclose(actual, expected)


@pytest.mark.parametrize("outer_bool", [False, True])
def test_roc_accessor(o, f_prob, outer_bool):
    bin_edges = np.linspace(0, 1, 6)
    actual = roc(o, f_prob, bin_edges=bin_edges)

    ds = xr.Dataset()
    ds["o"] = o
    ds["f_prob"] = f_prob
    if outer_bool:
        ds = ds.drop_vars("f_prob")
        expected = ds.xs.roc("o", f_prob, bin_edges=bin_edges)
    else:
        expected = ds.xs.roc("o", "f_prob", bin_edges=bin_edges)
    assert_allclose(actual, expected)



import numpy as np
import pytest
import xarray as xr
from xarray.tests import assert_allclose

from xskillscore.core.deterministic import (
    _preprocess_dims,
    _preprocess_weights,
    effective_sample_size,
    linslope,
    mae,
    mape,
    me,
    median_absolute_error,
    mse,
    pearson_r,
    pearson_r_eff_p_value,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_eff_p_value,
    spearman_r_p_value,
)
from xskillscore.core.np_deterministic import (
    _effective_sample_size,
    _linslope,
    _mae,
    _mape,
    _me,
    _median_absolute_error,
    _mse,
    _pearson_r,
    _pearson_r_eff_p_value,
    _pearson_r_p_value,
    _r2,
    _rmse,
    _smape,
    _spearman_r,
    _spearman_r_eff_p_value,
    _spearman_r_p_value,
)

correlation_metrics = [
    (linslope, _linslope),
    (pearson_r, _pearson_r),
    (r2, _r2),
    (pearson_r_p_value, _pearson_r_p_value),
    (pearson_r_eff_p_value, _pearson_r_eff_p_value),
    (spearman_r, _spearman_r),
    (spearman_r_p_value, _spearman_r_p_value),
    (spearman_r_eff_p_value, _spearman_r_eff_p_value),
    (effective_sample_size, _effective_sample_size),
]
correlation_metrics_names = [
    "linslope",
    "pearson_r",
    "r2",
    "pearson_r_p_value",
    "pearson_r_eff_p_value",
    "spearman_r",
    "spearman_r_p_value",
    "spearman_r_eff_p_value",
    "effective_sample_size",
]
distance_metrics = [
    (mse, _mse),
    (me, _me),
    (rmse, _rmse),
    (mae, _mae),
    (median_absolute_error, _median_absolute_error),
    (mape, _mape),
    (smape, _smape),
]
distance_metrics_names = [
    "mse",
    "me",
    "rmse",
    "mae",
    "median_absolute_error",
    "mape",
    "smape",
]

AXES = ("time", "lat", "lon", ["lat", "lon"], ["time", "lat", "lon"])

temporal_only_metrics = [
    pearson_r_eff_p_value,
    spearman_r_eff_p_value,
    effective_sample_size,
]


def adjust_weights(dim, weight_bool, weights):
    """
    Adjust the weights test data to only span the core dimension
    that the function is being applied over.
    """
    if weight_bool:
        drop_dims = [i for i in weights.dims if i not in dim]
        drop_dims = {k: 0 for k in drop_dims}
        return weights.isel(drop_dims)
    else:
        return None


@pytest.mark.parametrize("metrics", correlation_metrics, ids=correlation_metrics_names)
@pytest.mark.parametrize("dim", AXES)
@pytest.mark.parametrize("weight_bool", [True, False])
@pytest.mark.parametrize("skipna", [True, False])
@pytest.mark.parametrize("has_nan", [True, False])
def test_correlation_metrics_ufunc_same_np(
    a, b, dim, weight_bool, weights_cos_lat, metrics, skipna, has_nan
):
    """Test whether correlation metric for xarray functions (from
    deterministic.py) give save numerical results as for numpy functions (from
    np_deterministic.py)."""
    if has_nan:
        a[0] = np.nan
    # unpack metrics
    metric, _metric = metrics
    # Only apply over time dimension for effective p value.
    if (dim != "time") and (metric in temporal_only_metrics):
        dim = "time"
    # Generates subsetted weights to pass in as arg to main function and for
    # the numpy testing.
    _weights = adjust_weights(dim, weight_bool, weights_cos_lat)
    if metric in temporal_only_metrics:
        actual = metric(a, b, dim, skipna=skipna)
    else:
        actual = metric(a, b, dim, weights=_weights, skipna=skipna)
    # check that no chunks for no chunk inputs
    assert actual.chunks is None

    dim, _ = _preprocess_dims(dim, a)
    if len(dim) > 1:
        new_dim = "_".join(dim)
        _a = a.stack(**{new_dim: dim})
        _b = b.stack(**{new_dim: dim})
        if weight_bool:
            _weights = _weights.stack(**{new_dim: dim})
    else:
        new_dim = dim[0]
        _a = a
        _b = b
    _weights = _preprocess_weights(_a, dim, new_dim, _weights)

    # ensure _weights.values or None
    _weights = None if _weights is None else _weights.values

    axis = _a.dims.index(new_dim)
    if metric in temporal_only_metrics:
        res = _metric(_a.values, _b.values, axis, skipna=skipna)
    else:
        res = _metric(_a.values, _b.values, _weights, axis, skipna=skipna)
    expected = actual.copy()
    expected.values = res
    assert_allclose(actual, expected)


@pytest.mark.parametrize("metrics", correlation_metrics, ids=correlation_metrics_names)
@pytest.mark.parametrize("dim", AXES)
@pytest.mark.parametrize("weight_bool", [True, False])
@pytest.mark.parametrize("skipna", [True, False])
@pytest.mark.parametrize("has_nan", [True, False])
def test_correlation_metrics_daskda_same_npda(
    a_dask, b_dask, dim, weight_bool, weights_cos_lat_dask, metrics, skipna, has_nan
):
    """Test whether correlation metric for xarray functions can be lazy when
    chunked by using dask and give same results as DataArray with numpy array."""
    a = a_dask.copy()
    b = b_dask.copy()
    weights = weights_cos_lat_dask.copy()
    metric, _ = metrics
    if has_nan:
        a = a.load()
        a[0] = np.nan
        a = a.chunk()
    # Only apply over time dimension for effective p value.
    if (dim != "time") and (metric in temporal_only_metrics):
        dim = "time"
    # Generates subsetted weights to pass in as arg to main function
    _weights = adjust_weights(dim, weight_bool, weights)
    if metric in temporal_only_metrics:
        actual = metric(a, b, dim, skipna=skipna)
    else:
        actual = metric(a, b, dim, weights=_weights, skipna=skipna)
    if _weights is not None:
        _weights = _weights.load()
    if metric in temporal_only_metrics:
        expected = metric(a.load(), b.load(), dim, skipna=skipna)
    else:
        expected = metric(a.load(), b.load(), dim, weights=_weights, skipna=skipna)
    assert_allclose(actual, expected)


@pytest.mark.parametrize("metrics", distance_metrics, ids=distance_metrics_names)
@pytest.mark.parametrize("dim", AXES)
@pytest.mark.parametrize("weight_bool", [True, False])
@pytest.mark.parametrize("skipna", [True, False])
@pytest.mark.parametrize("has_nan", [True, False])
def test_distance_metrics_ufunc_same_np(
    a, b, dim, weight_bool, weights_cos_lat, metrics, skipna, has_nan
):
    """Test whether distance-based metric for xarray functions (from
    deterministic.py) give save numerical results as for numpy functions (from
    np_deterministic.py)."""
    if has_nan:
        a[0] = np.nan
    # unpack metrics
    metric, _metric = metrics
    # Generates subsetted weights to pass in as arg to main function and for
    # the numpy testing.
    weights = adjust_weights(dim, weight_bool, weights_cos_lat)
    # median absolute error has no weights argument
    if metric is median_absolute_error:
        actual = metric(a, b, dim, skipna=skipna)
    else:
        actual = metric(a, b, dim, weights=weights, skipna=skipna)
    assert actual.chunks is None

    dim, axis = _preprocess_dims(dim, a)
    _a = a
    _b = b
    _weights = _preprocess_weights(_a, dim, dim, weights)
    axis = tuple(a.dims.index(d) for d in dim)
    if metric is median_absolute_error:
        res = _metric(_a.values, _b.values, axis, skipna=skipna)
    else:
        # ensure _weights.values or None
        _weights = None if _weights is None else _weights.values
        res = _metric(_a.values, _b.values, _weights, axis, skipna=skipna)
    expected = actual.copy()
    expected.values = res
    assert_allclose(actual, expected)


@pytest.mark.parametrize("metrics", distance_metrics, ids=distance_metrics_names)
@pytest.mark.parametrize("dim", AXES)
@pytest.mark.parametrize("weight_bool", [True, False])
@pytest.mark.parametrize("skipna", [True, False])
@pytest.mark.parametrize("has_nan", [True, False])
def test_distance_metrics_daskda_same_npda(
    a_dask, b_dask, dim, weight_bool, weights_cos_lat_dask, metrics, skipna, has_nan
):
    """Test whether distance metric for xarray functions can be lazy when
    chunked by using dask and give same results as DataArray with numpy array."""
    a = a_dask.copy()
    b = b_dask.copy()
    weights = weights_cos_lat_dask.copy()
    metric, _ = metrics
    if has_nan:
        a = a.load()
        a[0] = np.nan
        a = a.chunk()
    # Generates subsetted weights to pass in as arg to main function
    _weights = adjust_weights(dim, weight_bool, weights)
    if metric is median_absolute_error:
        actual = metric(a, b, dim, skipna=skipna)
    else:
        actual = metric(a, b, dim, weights=_weights, skipna=skipna)
    if metric is median_absolute_error:
        expected = metric(a.load(), b.load(), dim, skipna=skipna)
    else:
        expected = metric(a.load(), b.load(), dim, weights=_weights, skipna=skipna)
    assert_allclose(actual, expected)


@pytest.mark.parametrize("dim", AXES)
@pytest.mark.parametrize("metric", [smape], ids=["smape"])
def test_percentage_metric_in_interval_0_1(a, b, dim, metric):
    """Test smape to be within bounds."""
    res = metric(a, b, dim)
    assert not (res < 0).any()
    assert not (res > 1).any()
    assert not res.isnull().any()


def test_pearson_r_p_value_skipna(a, b_rand_nan):
    """Test whether NaNs sprinkled in array will NOT yield all NaNs."""
    res = pearson_r_p_value(a, b_rand_nan, ["lat", "lon"], skipna=True)
    assert not np.isnan(res).all()


# def test_pearson_r_eff_p_value_skipna(a, b_nan):
#     """Test whether NaNs sprinkled in array will NOT yield all NaNs."""
#     res = pearson_r_eff_p_value(a, b_nan, ["lat", "lon"], skipna=True)
#     assert not np.isnan(res).all()


def test_pearson_r_integer():
    """Test whether arrays as integers work."""
    da = xr.DataArray([0, 1, 2], dims=["time"])
    assert pearson_r(da, da, dim="time") == 1


@pytest.mark.parametrize(
    "metrics",
    correlation_metrics + distance_metrics,
    ids=correlation_metrics_names + distance_metrics_names,
)
@pytest.mark.parametrize("keep_attrs", [True, False])
def test_keep_attrs(a, b, metrics, keep_attrs):
    """Test keep_attrs for all metrics."""
    metric, _metric = metrics
    # ths tests only copying attrs from a
    res = metric(a, b, "time", keep_attrs=keep_attrs)
    if keep_attrs:
        assert res.attrs == a.attrs
    else:
        assert res.attrs == {}
    da = xr.DataArray([0, 1, 2], dims=["time"])
    assert pearson_r(da, da, dim="time") == 1


@pytest.mark.parametrize(
    "metrics",
    correlation_metrics + distance_metrics,
    ids=correlation_metrics_names + distance_metrics_names,
)
def test_dim_None(a, b, metrics):
    """Test that `dim=None` reduces all dimensions as xr.mean(dim=None) and fails for
    effective metrics."""
    metric, _metric = metrics
    if metric in [effective_sample_size, spearman_r_eff_p_value, pearson_r_eff_p_value]:
        with pytest.raises(ValueError) as excinfo:
            metric(a, b, dim=None)
        assert (
            "Effective sample size should only be applied to a singular time dimension."
            in str(excinfo.value)
        )
    else:
        metric, _metric = metrics
        res = metric(a, b, dim=None)
        assert len(res.dims) == 0, print(res.dims)


@pytest.mark.parametrize(
    "metrics",
    correlation_metrics + distance_metrics,
    ids=correlation_metrics_names + distance_metrics_names,
)
def test_dim_empty_list(a, b, metrics):
    """Test that `dim=[]` reduces no dimensions as xr.mean(dim=[]) and fails for
    correlation metrics."""
    if metrics in correlation_metrics:
        metric, _metric = metrics
        with pytest.raises(ValueError) as excinfo:
            metric(a, b, dim=[])
        assert "requires `dim` not being empty, found dim" in str(excinfo.value)
    elif metrics in distance_metrics:
        metric, _metric = metrics
        res = metric(a, b, dim=[])
        assert len(res.dims) == len(a.dims), print(res.dims)


@pytest.mark.parametrize(
    "metrics",
    correlation_metrics + distance_metrics,
    ids=correlation_metrics_names + distance_metrics_names,
)
def test_correlation_broadcasts(a, b, metrics):
    """Test whether correlation metric broadcasts dimensions other than dim."""
    # unpack metrics
    metric, _metric = metrics
    metric(a, b.isel(lat=0), dim="time")
    metric(a, b.isel(lat=[0]), dim="time")
    b_changed_coords = b.isel(lat=[0]).assign_coords(lat=[123])
    if (
        "eff" not in metric.__name__
    ):  # effective metrics require to be applied over time
        with pytest.raises(ValueError) as e:
            metric(a, b_changed_coords, dim="lat")
        assert "indexes along dimension" in str(e.value)



import numpy as np
import pytest

from xskillscore.core.deterministic import (
    linslope,
    mae,
    mape,
    me,
    median_absolute_error,
    mse,
    pearson_r,
    pearson_r_p_value,
    r2,
    rmse,
    smape,
    spearman_r,
    spearman_r_p_value,
)

METRICS = [
    linslope,
    mae,
    mse,
    median_absolute_error,
    mape,
    smape,
    me,
    rmse,
    pearson_r,
    pearson_r_p_value,
    spearman_r,
    spearman_r_p_value,
    r2,
]


def mask_data(da):
    """
    Masks sample data arbitrarily like a block of land.
    """
    da.data[:, 1:3, 1:3] = np.nan
    da.data[0:2, 0, 0] = np.nan
    return da


@pytest.mark.parametrize("metric", METRICS)
def test_single_grid_cell_matches_individual_time_series(a, b, metric):
    """Test that a single grid cell result from a gridded dataset matches the
    result for just passing through the individual time series"""
    for x in range(a["lon"].size):
        for y in range(a["lat"].size):
            ts_a = a.isel(lat=y, lon=x)
            ts_b = b.isel(lat=y, lon=x)
            gridded_res = metric(a, b, "time").isel(lat=y, lon=x)
            ts_res = metric(ts_a, ts_b, "time")
            assert np.allclose(gridded_res, ts_res)


@pytest.mark.parametrize("metric", METRICS)
def test_single_grid_cell_matches_individual_time_series_nans(a, b, metric):
    """Test that a single grid cell result from a gridded dataset with nans matches the
    result for just passing through the individual time series"""
    a_masked = mask_data(a)
    b_masked = mask_data(b)
    for x in range(a_masked["lon"].size):
        for y in range(a_masked["lat"].size):
            ts_a_masked = a_masked.isel(lat=y, lon=x)
            ts_b_masked = b_masked.isel(lat=y, lon=x)
            gridded_res = metric(a_masked, b_masked, "time", skipna=True).isel(
                lat=y, lon=x
            )
            ts_res = metric(ts_a_masked, ts_b_masked, "time", skipna=True)
            assert np.allclose(gridded_res, ts_res, equal_nan=True)
